import math
import numpy as np
from datetime import datetime, timedelta
from enum import Enum
from scipy.signal.windows import hamming, tukey
import scipy.signal
import statsmodels.api as sm
import colorednoise as cn


class windowing_methods(Enum):
    """
    The available windowing methods for the waveform
    """

    hanning = 1
    hamming = 2
    tukey = 3
    rectangular = 4


class trimming_methods(Enum):
    """
    Trimming can be accomplished with either the samples or times. This enumeration defines whether to use the time to
    calculate the sample or just provide the samples.
    """

    samples = 1
    times = 2


class scaling_method(Enum):
    """
    In scaling the waveform we can apply the level changes in either decibels or linear values. This will determine how
    the interface scales the signal when manipulating the sample magnitudes.
    """

    linear = 1
    logarithmic = 2


class weighting_function(Enum):
    """
    This class provides the options on how to weight the calculation of the overall level values
    """

    unweighted = 0
    a_weighted = 1
    c_weighted = 2


class correlation_mode(Enum):
    """
    This class defines the various modes for the cross-correlation function.
    """

    valid = 0
    full = 1
    same = 2


class noise_color(Enum):
    white = 0
    pink = 1
    brown = 2


class Waveform:
    """
    This is a generic base class that contains the start time, samples and sample rate for a waveform.  Some limited
    operations exist within this class for manipulation of the base data within the class.

    #   TODO: Add function to overload the __add__ operator so that two waveforms can be combined appropriately
    #   TODO: Add convolution function
    #   TODO: Connect the WaveFile Start time to the Waveform start time
    #   TODO: Add function for Leq calculation
    #   TODO: Add function for SEL calculation
    #   TODO: add band pass filter
    #   TODO: Add filter array function
    #   TODO: create function for calculating the engineering scale factor from a waveform
    Remarks
    2022-05-11 - FSM - added the function to determine whether the waveform is a calibration signal or not.
    """

    def __init__(self, pressures, sample_rate, start_time, remove_dc_offset: bool = True):
        """
        Default constructor
        :param pressures: float, array-like - the list of pressure values
        :param sample_rate: float - the number of samples per second
        :param start_time: float or datetime - the time of the first sample
        """

        self._samples = pressures
        if remove_dc_offset:
            self._samples -= np.mean(self._samples)
        self.fs = sample_rate
        self.time0 = start_time
        self._forward_coefficients = None
        self._reverse_coefficients = None

        self._coefficient_count = 12
        self._hop_size_seconds = 0.0029
        self._window_size_seconds = 0.0232
        self._cutoff_frequency = 5
        self._centroid_threshold = 0.15
        self._effective_duration_threshold = 0.4

        self._signal_envelope = None
        self._normal_signal_envelope = None
        self._log_attack = None
        self._increase = None
        self._decrease = None
        self._Addresses = None
        self._amplitude_modulation = None
        self._frequency_modulation = None
        self._auto_correlation_coefficients = None
        self._zero_cross_rate = None
        self._temporal_centroid = None
        self._effective_duration = None
        self._temporal_feature_times = None

    # ---------------------- Collection of properties - this is both getters and setters -------------------------------

    @property
    def duration(self):
        """
        Determine the duration of the waveform by examining the number of samples and the sample rate
        :return: float - the total number of seconds within the waveform
        """
        return float(len(self._samples)) / self.fs

    @property
    def end_time(self):
        """
        Determine the end time - if the start time was a datetime, then this returns a datetime.  Otherwise a floating
        point value is returned
        :return: float or datetime - the end of the file
        """
        if isinstance(self.time0, datetime):
            return self.time0 + timedelta(seconds=self.duration)
        else:
            return self.time0 + self.duration

    @property
    def samples(self):
        """
        The actual pressure waveform
        :return: float, array-like - the collection of waveform data
        """
        return self._samples

    @samples.setter
    def samples(self, array):
        self._samples = array

    @property
    def sample_rate(self):
        """
        The number of samples per second to define the waveform.
        :return: float - the number of samples per second
        """
        return self.fs

    @sample_rate.setter
    def sample_rate(self, value):
        self.fs = value

    @property
    def start_time(self):
        """
        The time of the first sample
        :return: float or datetime - the time of the first sample
        """

        return self.time0

    @start_time.setter
    def start_time(self, value):
        self.time0 = value

    @property
    def forward_coefficients(self):
        return self._forward_coefficients

    @property
    def reverse_coefficients(self):
        return self._reverse_coefficients

    @property
    def times(self):
        """
        This determines the time past midnight for the start of the audio and returns a series of times for each sample
        :return: float, array-like - the sample times for each element of the samples array
        """

        if isinstance(self.start_time, datetime):
            t0 = (60 * (60 * self.start_time.hour + self.start_time.minute) + self.start_time.second +
                  self.start_time.microsecond * 1e-6)
        else:
            t0 = self.start_time

        return np.arange(0, len(self.samples)) / self.sample_rate + t0

    @property
    def effective_duration_threshold(self):
        return self._effective_duration_threshold

    @property
    def centroid_threshold(self):
        return self._centroid_threshold

    @property
    def cutoff_frequency(self):
        return self._cutoff_frequency

    @cutoff_frequency.setter
    def cutoff_frequency(self, value):
        self._cutoff_frequency = value

    @property
    def window_size_seconds(self):
        return self._window_size_seconds

    @window_size_seconds.setter
    def window_size_seconds(self, value):
        self._window_size_seconds = value

    @property
    def hop_size_seconds(self):
        return self._hop_size_seconds

    @hop_size_seconds.setter
    def hop_size_seconds(self, value):
        self._hop_size_seconds = value

    @property
    def window_size_samples(self):
        return int(np.round(self.window_size_seconds * self.sample_rate))

    @property
    def hop_size_samples(self):
        return int(round(self.hop_size_seconds * self.sample_rate))

    @property
    def coefficient_count(self):
        """
        The number of coefficients to generate for the available data
        """

        return self._coefficient_count

    @coefficient_count.setter
    def coefficient_count(self, value):
        """
        Set the number of coefficients for the analysis
        """

        self._coefficient_count = value

    @property
    def attack(self):
        if self._Addresses is None:
            self._calculate_signal_envelope()
            self._log_attack, self._increase, self._decrease, self._Addresses = self.calculate_log_attack()

        return self._Addresses[0]

    @property
    def decrease(self):
        if self._Addresses is None:
            self._calculate_signal_envelope()
            self._log_attack, self._increase, self._decrease, self._Addresses = self.calculate_log_attack()

        return self._Addresses[1]

    @property
    def release(self):
        if self._Addresses is None:
            self._calculate_signal_envelope()
            self._log_attack, self._increase, self._decrease, self._Addresses = self.calculate_log_attack()

        return self._Addresses[4]

    @property
    def log_attack(self):
        """
        The log-attack-time is simply defined as LAT = log_10(t[-1]-t[0])
        """
        if self._Addresses is None:
            self._calculate_signal_envelope()
            self._log_attack, self._increase, self._decrease, self._Addresses = self.calculate_log_attack()

        return self._log_attack

    @property
    def attack_slope(self):
        """
        The attack slope is defined as the average temporal slope of the energy during the attack segment. We compute
        the local slopes of the energy corresponding to each effort w_i. We then compute a weighted average of the
        slopes. The weights are chosen in order to emphasize slope values in the middle of the attack (the weights are
        the values of a Gaussian function centered around the threshold = 50% and with a standard-deviation of 0.5).
        """
        if self._Addresses is None:
            self._calculate_signal_envelope()
            self._log_attack, self._increase, self._decrease, self._Addresses = self.calculate_log_attack()

        return self._increase

    @property
    def decrease_slope(self):
        """
        The temporal decrease is a measure of the rate of decrease of the signal energy. It distinguishes non-sustained
        (e.g. percussive, pizzicato) sounds from sustained sounds. Its calculation is based on a decreasing exponential
        model of the energy envelope starting from it maximum.
        """
        if self._Addresses is None:
            self._calculate_signal_envelope()
            self._log_attack, self._increase, self._decrease, self._Addresses = self.calculate_log_attack()

        return self._decrease

    @property
    def temporal_centroid(self):
        """
        The temporal centroid is the center of gravity of the energy envelope. It distinguishes percussive from
        sustained sounds. It has been proven to be a perceptually important descriptor (Peeters et al., 2000).
        """
        if self._temporal_centroid is None:
            self._calculate_signal_envelope()
            self._temporal_centroid = self.calculate_temporal_centroid()

        return self._temporal_centroid

    @property
    def effective_duration(self):
        """
        The effective duration is a measure intended to reflect the perceived duration of the signal. It distinguishes
        percussive sounds from sustained sounds but depends on the event duration. It is approximated by the time the
        energy envelop is above a given threshold. After many empirical tests, we have set this threshold to 40%
        """
        if self._effective_duration is None:
            self._calculate_signal_envelope()
            self._effective_duration = self.calculate_effective_duration()

        return self._effective_duration

    @property
    def amplitude_modulation(self):
        if self._amplitude_modulation is None:
            self._calculate_signal_envelope()
            self._log_attack, self._increase, self._decrease, self._Addresses = self.calculate_log_attack()
            self._frequency_modulation, self._amplitude_modulation = self.calculate_modulation()
        return self._amplitude_modulation

    @property
    def frequency_modulation(self):
        if self._frequency_modulation is None:
            self._calculate_signal_envelope()
            self._log_attack, self._increase, self._decrease, self._Addresses = self.calculate_log_attack()
            self._frequency_modulation, self._amplitude_modulation = self.calculate_modulation()
        return self._frequency_modulation

    @property
    def auto_correlation(self):
        if self._auto_correlation_coefficients is None:
            self._temporal_feature_times, self._auto_correlation_coefficients, self._zero_cross_rate = \
                self.instantaneous_temporal_features()
        return self._auto_correlation_coefficients

    @property
    def zero_crossing_rate(self):
        if self._zero_cross_rate is None:
            self._temporal_feature_times, self._auto_correlation_coefficients, self._zero_cross_rate = \
                self.instantaneous_temporal_features()
        return self._zero_cross_rate

    @property
    def temporal_feature_times(self):
        if self._temporal_feature_times is None:
            self._temporal_feature_times, self._auto_correlation_coefficients, self._zero_cross_rate = \
                self.instantaneous_temporal_features()
        return self._temporal_feature_times

    @property
    def signal_envelope(self):
        if self._signal_envelope is None:
            self._calculate_signal_envelope()

        return self._signal_envelope

    @property
    def normal_signal_envelope(self):
        if self._normal_signal_envelope is None:
            self._calculate_signal_envelope()

        return self._normal_signal_envelope

    @property
    def loudness(self):
        from mosqito.sq_metrics import loudness_zwst_perseg
        return loudness_zwst_perseg(signal=self.samples, fs=self.sample_rate)[0]

    @property
    def roughness(self):
        from mosqito.sq_metrics import roughness_dw

        return roughness_dw(signal=self.samples, fs=self.sample_rate)[0]

    @property
    def sharpness(self):
        from mosqito.sq_metrics import sharpness_din_perseg

        return sharpness_din_perseg(signal=self.samples, fs=self.sample_rate)[0]

    @property
    def is_mono(self):
        return len(self.samples.shape) == 1

    def is_clipped(self):
        """
        This function attempts to determine whether there is clipping in the acoustic data represented in this waveform.
        """

        return false

    # ------------------ Static functions for the calculation of filter shapes and timbre features ---------------------

    @staticmethod
    def AC_Filter_Design(fs):
        """
        AC_Filter_Design.py

        Created on Mon Oct 18 19:27:36 2021

        @author: Conner Campbell, Ball Aerospace

        Description
        ----------
        Coeff_A, Coeff_C = AC_Filter_Design(fs)

        returns Ba, Aa, and Bc, Ac which are arrays of IRIR filter
        coefficients for A and C-weighting.  fs is the sampling
        rate in Hz.

        This progam is a recreation of adsgn and cdsgn
        by Christophe Couvreur, see	Matlab FEX ID 69.


        Parameters
        ----------
        fs : double
            sampling rate in Hz

        Returns
        -------

        Coeff_A: list
            List of two numpy arrays, feedforward and feedback filter
            coeffecients for A-weighting filter. Form of lits is [Ba,Aa]

        Coeff_c: list
            List of two numpy arrays, feedforward and feedback filter
            coeffecients for C-weighting filter. Form of lits is [Bc,Ac]

        Code Dependencies
        -------
        This program requires the following python packages:
        scipy.signal, numpy

        References
        -------
        IEC/CD 1672: Electroacoustics-Sound Level Meters, Nov. 1996.

        ANSI S1.4: Specifications for Sound Level Meters, 1983.

        ACdsgn.m: Christophe Couvreur, Faculte Polytechnique de Mons (Belgium)
        couvreur@thor.fpms.ac.be
        """

        # Define filter poles for A/C weight IIR filter according to IEC/CD 1672

        f1 = 20.598997
        f2 = 107.65265
        f3 = 737.86223
        f4 = 12194.217
        A1000 = 1.9997
        C1000 = 0.0619
        pi = np.pi

        # Calculate denominator and numerator of filter tranfser functions

        coef1 = (2 * pi * f4) ** 2 * (10 ** (C1000 / 20))
        coef2 = (2 * pi * f4) ** 2 * (10 ** (A1000 / 20))

        Num1 = np.array([coef1, 0.0])
        Den1 = np.array([1, 4 * pi * f4, (2 * pi * f4) ** 2])

        Num2 = np.array([1, 0.0])
        Den2 = np.array([1, 4 * pi * f1, (2 * pi * f1) ** 2])

        Num3 = np.array([coef2 / coef1, 0.0, 0.0])
        Den3 = scipy.signal.convolve(np.array([1, 2 * pi * f2]).T, (np.array([1, 2 * pi * f3])))

        # Use scipy.signal.bilinear function to get numerator and denominator of
        # the transformed digital filter transfer functions.

        B1, A1 = scipy.signal.bilinear(Num1, Den1, fs)
        B2, A2 = scipy.signal.bilinear(Num2, Den2, fs)
        B3, A3 = scipy.signal.bilinear(Num3, Den3, fs)

        Ac = scipy.signal.convolve(A1, A2)
        Aa = scipy.signal.convolve(Ac, A3)

        Bc = scipy.signal.convolve(B1, B2)
        Ba = scipy.signal.convolve(Bc, B3)

        Coeff_A = [Ba, Aa]
        Coeff_C = [Bc, Ac]
        return Coeff_A, Coeff_C

    @staticmethod
    def detect_local_extrema(input_v, lag_n):
        """
        This will detect the local maxima of the vector on the interval [n-lag_n:n+lag_n]

        Parameters
        ----------
        input_v : double array-like
            This is the input vector that we are examining to determine the local maxima
        lag_n : double, integer
            This is the number of samples that we are examining within the input_v to determine the local maximum

        Returns
        -------
        pos_max_v : double, array-like
            The locations of the local maxima
        """

        do_affiche = 0
        lag2_n = 4
        seuil = 0

        L_n = len(input_v)

        pos_cand_v = np.where(np.diff(np.sign(np.diff(input_v))) < 0)[0]
        pos_cand_v += 1

        pos_max_v = np.zeros((len(pos_cand_v),))

        for i in range(len(pos_cand_v)):
            pos = pos_cand_v[i]

            if (pos > lag_n) & (pos <= L_n - lag_n):
                tmp = input_v[pos - lag_n:pos + lag2_n]
                position = np.argmax(tmp)

                position = position + pos - lag_n - 1

                if (pos - lag2_n > 0) & (pos + lag2_n < L_n + 1):
                    tmp2 = input_v[pos - lag2_n:pos + lag2_n]

                    if (position == pos) & (input_v[position] > seuil * np.mean(tmp2)):
                        pos_max_v[i] = pos

        return pos_max_v

    @staticmethod
    def next_pow2(x: int):
        n = np.log2(x)
        return 2 ** (np.floor(n) + 1)

    @staticmethod
    def gps_audio_signal_converter(filename: str):
        """
        This function will convert the data from an audio file that captured the GPS signal and converted it to an
        audio channel.

        This function is adapted from the Python code delivered as part of the ESCAPE 2018 dataset (gps_wav.py)

        Parameters
        ----------
        filename: str - the full path to the audio file
        """
        import wave
        import struct
        import scipy.signal
        import re

        #   Initialize some parsing variables
        start_search = 2
        win = 1000
        thrE = 1000
        thr = 2500
        gap = 99

        #   Load the wave data
        wave_object = wave.open(filename, 'rb')
        channel_count = wave_object.getnchannels()
        bit_count = wave_object.getsampwidth()
        sample_rate = wave_object.getframerate()
        frame_count = wave_object.getnframes()
        temp_data = wave_object.readframes(frame_count)
        data = struct.unpack("<" + str(frame_count) + "h", temp_data)

        #   Filter the data with a high-pass filter to remove any DC offsets
        data = scipy.signal.lfilter([1.0, -1.0], 1.0, data)

        #   Find first complete GPS data burst
        energy = []
        filtered_data = []
        for i in range(0, start_search * sample_rate, win):
            energy = np.sum(np.square(data[i:i + win])) / win
            if energy < thrE:
                offset = i + win
                filtered_data.extend(data[offset:frame_count])
                frame_count = len(filtered_data)
                break

        # Detect pulse state transitions in audio signal
        state = True
        detect = []
        ind = []
        indPlus = []
        for i in range(1, np.min([250000, frame_count])):
            if (filtered_data[i] >= filtered_data[i + 1]) & (filtered_data[i] > filtered_data[i - 1]):
                if filtered_data[i] > thr:
                    detect.append(state)
                    state = not state
                    ind.append(i)
                    indPlus.append(i)
            elif (filtered_data[i] <= filtered_data[i + 1]) & (filtered_data[i] < filtered_data[i - 1]):
                if filtered_data[i] < -thr:
                    detect.append(state)
                    state = not state
                    ind.append(i)

        # Extract bit sequence and corresponding sample indices
        bits = []
        index = []
        index.append(ind[0])
        for i in range(len(ind) - 1):
            if (ind[i + 1] - ind[i]) > gap:
                bits.append(1)
                index.append(ind[i + 1])
                # print(ind[i+1])
                continue
            if detect[i]:
                bits.extend(np.zeros(int(round((ind[i + 1] - ind[i]) / 10.0)), dtype=np.int))
            else:
                bits.extend(np.ones(int(round((ind[i + 1] - ind[i]) / 10.0)), dtype=np.int))

        #   Extract NMEA 0183 sentences from bit sequence and write to file
        s = ''
        sent = ''
        cnt = 0
        for i in range(int(len(bits) / 10)):
            word = s.join(map(str, bits[cnt:cnt + 10]))
            cnt = cnt + 10
            if bool(int(word[0])) or bool(int(word[8])) or not (bool(int(word[9]))):
                # continue
                wordE = []
                for k in range(10):
                    wordE.append(str(int(not (bool(int(word[k]))))))
                word = s.join(wordE)
            sent = sent + chr(int('0b' + word[8:0:-1], 2))

        elements = sent.split('\r\n')

        for i in range(len(elements)):
            if elements[i].split(',')[0] == "$GPRMC":
                first_rmc = elements[i]
                break

        # Find start of 1st complete GPS data bursts
        indBurst = indPlus[0] + offset
        elements = first_rmc.split(',')
        hour = int(elements[1][:2])
        minute = int(elements[1][2:4])
        start_time = (60 * (60 * int(elements[1][:2]) + int(elements[1][2:4])) + float(elements[1][4:]) - indBurst /
                      sample_rate)
        date = datetime(int(elements[9][4:]) + 2000, int(elements[9][2:4]), int(elements[9][:2])) + \
               timedelta(seconds=start_time)

        return date

    @staticmethod
    def generate_tone(frequency: float = 100, sample_rate: float = 48000, duration: float = 1.0,
                      amplitude_db: float = 94):
        """
        This function generates a sine wave tone function with the specific frequency and duration specified in the
        argument list.

        Parameters
        ----------
        frequency: float, default: 100 - the linear frequency of the waveform
        sample_rate: float, default: 48000 - the number of samples per second
        duration: float, default: 1.0 - the total number of seconds in the waveform
        amplitude_db: float, default:94, this is the RMS amplitude of the waveform

        Returns
        -------
        A waveform this the generated data.
        """

        amplitude_rms = 10 ** (amplitude_db / 20) * 2e-5
        x = np.arange(0, duration, 1 / sample_rate)
        y = amplitude_rms * np.sqrt(2) * np.sin(2 * np.pi * frequency * x)

        return Waveform(y, sample_rate, 0)

    @staticmethod
    def generate_noise(sample_rate: float = 48000, duration: float = 1.0, amplitude_db: float = 94,
                       type=noise_color.pink):
        samples = cn.powerlaw_psd_gaussian(type.value, int(np.floor(duration * sample_rate)))

        wfm = Waveform(samples, sample_rate, 0)
        scaling = wfm.overall_level() - amplitude_db
        wfm.scale_signal(scaling, True, scaling_method.logarithmic)

        return wfm

    @staticmethod
    def irig_converter(signal):
        """
        Compute the time of the signal using the IRIG-B format as reference.

        Parameters
        ----------
        signal : double, array-like

        Returns
        -------
        datetime object for the start of the sample
        """

        irig = signal * 30.0 / np.max(signal)
        si = np.sign(irig - np.mean(irig))

        dsi = np.diff(si)

        rise = np.where(dsi == 2)[0]
        fall = np.where(dsi == -2)[0]

        if np.min(fall) < np.min(rise):
            fall = fall[1:]

        if np.max(rise) > np.max(fall):
            rise = rise[:-1]

        rf = np.stack([rise, fall]).transpose()

        index = np.round(np.mean(rf, axis=1, dtype='int'))
        top = irig[index]
        top2 = (top > 20) * 30 + (top < 20) * 10 - 10

        p0pr = np.array([30, 30, 30, 30, 30, 30, 30, 30, 10, 10, 30, 30, 30, 30, 30, 30, 30, 30]) - 10

        #   Locate this sequence in the top2 array

        pr = list()
        for i in range(len(top2) - len(p0pr)):
            located = True
            for j in range(len(p0pr)):
                if top2[i + j] != p0pr[j]:
                    located = False
                    break
            if located:
                pr.append(i + 10)

        prrise = rise[pr]
        sps = np.mean(np.diff(prrise))

        carr = np.mean(np.diff(pr))

        seconds = np.zeros((len(pr) - 1,))
        minutes = np.zeros((len(pr) - 1,))
        hours = np.zeros((len(pr) - 1,))
        day_of_year = np.zeros((len(pr) - 1,))
        dt = np.zeros((len(pr) - 1,))

        for j in range(len(pr) - 1):

            start_index = int(pr[j] + 0.01 * carr)
            stop_index = int(pr[j] + 0.01 * 2 * carr)

            values = np.array([1, 2, 4, 8, 0, 10, 20, 40])
            mask = np.zeros((len(values),))

            for i in range(len(mask)):
                if np.sum(top2[start_index:stop_index]) > 70:
                    mask[i] = 1

                start_index += 10
                stop_index += 10

            seconds[j] = np.sum(values * mask)

            start_index = int(pr[j] + 0.01 * carr * 10)
            stop_index = int(pr[j] + 0.01 * 11 * carr)

            values = np.array([1, 2, 4, 8, 0, 10, 20, 40])
            mask = np.zeros((len(values),))

            for i in range(len(mask)):
                if np.sum(top2[start_index:stop_index]) > 70:
                    mask[i] = 1

                start_index += 10
                stop_index += 10

            minutes[j] = np.sum(values * mask)

            start_index = int(pr[j] + 0.01 * carr * 20)
            stop_index = int(pr[j] + 0.01 * 21 * carr)

            values = np.array([1, 2, 4, 8, 0, 10, 20])
            mask = np.zeros((len(values),))

            for i in range(len(mask)):
                if np.sum(top2[start_index:stop_index]) > 70:
                    mask[i] = 1

                start_index += 10
                stop_index += 10

            hours[j] = np.sum(values * mask)

            start_index = int(pr[j] + 0.01 * carr * 30)
            stop_index = int(pr[j] + 0.01 * 31 * carr)

            values = np.array([1, 2, 4, 8, 0, 10, 20, 40, 80, 0, 100, 200])
            mask = np.zeros((len(values),))

            for i in range(len(mask)):
                if np.sum(top2[start_index:stop_index]) > 70:
                    mask[i] = 1

                start_index += 10
                stop_index += 10

            day_of_year[j] = np.sum(values * mask)

            #   Determine the linear adjustment for the zero cross over not occurring right on the sample

            dt[j] = (np.interp(0, [irig[prrise[j]], irig[prrise[j] + 1]], [prrise[j], prrise[j] + 1]) - prrise[j]) / sps

        #   Compute the time past midnight

        times = 60 * (60 * hours + minutes) + seconds - dt

        day_of_year = np.mean(day_of_year)

        index = np.arange(0, len(irig))
        timevector = times[0] + (index - prrise[0]) / sps

        return times[0] - prrise[0] / sps, day_of_year

    @staticmethod
    def irig_converter_for_arc(data, fs):
        """
        The timecode generators present at the Aeroacoustic Research Complex (ARC) produce the signal that defines the IRIG-
        B timecode differently. The previous methods do not return the correct information for the ARC data.


        """
        #   Find the index of the first minimum - which is assumed to occur within the first second of the waveform

        index = np.where(data[:fs] == np.min(data[:fs]))[0][0]

        #   The IRIG-B waveform is an amplitude modulated 1 kHz sine wave.  So we need to know the number of samples
        #   for a single period of the waveform

        frequency = 1000
        period = 1 / frequency
        period_samples = period * fs

        #   Now we can get the first set of signals and determine the amplitude for the minima

        amplitudes = np.zeros((3000,))

        for i in range(3000):
            amplitudes[i] = data[int(index + i * period_samples)]

            if index + (i + 1) * period_samples >= len(data):
                break

        maximum_index = i

        #   Scale by the smallest value in this array

        amplitudes /= np.min(amplitudes)

        #   Convert this to a binary array that will be used for the determination of the bit-wise elements of the
        #   signal

        binary_array = np.zeros((maximum_index,))

        for i in range(maximum_index):
            if amplitudes[i] >= 0.8:
                binary_array[i] = 1

        #   Now the start of a signal must be with a signal that is 8, so we need to locate the first 8 within the
        #   summation of the binary signal.

        first_eight = -1
        for i in range(3000):
            elements = binary_array[i:i + 10]

            if np.sum(elements) == 8 and binary_array[i] == 1 and binary_array[i + 9] == 0:
                first_eight = i
                break

        if first_eight < 0:
            return None, None

        #   Now that we have determined the location of the first 8, we can begin to sum the binary array in
        #   sections of 10 elements at a time

        summed_array = np.zeros((250,))

        for i in range(250):
            idx = first_eight + i * 10
            summed_array[i] = np.sum(binary_array[idx:idx + 10])

        #   Now we must find the first double-8 as this marks the beginning of the time code definition

        first_double_eight = -1
        for i in range(250):
            if summed_array[i] == 8 and summed_array[i + 1] == 8:
                first_double_eight = i + 1
                break

        if first_double_eight < 0:
            return None, None

        #   Now from this we need to extract the various parts of the time code

        timecode_elements = summed_array[first_double_eight:first_double_eight + 100]

        #   Get the hours

        hour_elements = timecode_elements[20:30]
        hour_elements[np.where(hour_elements < 5)[0]] = 0
        hour_elements[np.where(hour_elements >= 5)[0]] = 1

        weights = np.array([1, 2, 4, 8, 0, 10, 20, 0, 0, 0])

        hour = np.sum(hour_elements * weights)

        #   Get the minutes

        minute_elements = timecode_elements[10:20]
        minute_elements[np.where(minute_elements < 5)[0]] = 0
        minute_elements[np.where(minute_elements == 5)[0]] = 1
        weights = np.array([1, 2, 4, 8, 0, 10, 20, 40, 0, 0])

        minutes = np.sum(minute_elements * weights)

        #   Next the seconds

        seconds_elements = timecode_elements[:10]
        seconds_elements[np.where(seconds_elements < 5)[0]] = 0
        seconds_elements[np.where(seconds_elements == 5)[0]] = 1
        weights = np.array([0, 1, 2, 4, 8, 0, 10, 20, 40, 0])

        seconds = np.sum(seconds_elements * weights)

        #   And finally we get the julian date of the time code

        date_elements = timecode_elements[30:42]
        date_elements[np.where(date_elements != 5)[0]] = 0
        date_elements[np.where(date_elements == 5)[0]] = 1
        weights = np.array([1, 2, 4, 8, 0, 10, 20, 40, 80, 0, 100, 200])

        julian_date = np.sum(weights * date_elements)

        #   Now we know that the time code was not at the very beginning of the signal, so let's go ahead and
        #   determine the time offset to the beginning of the file.

        tpm = 60 * (60 * hour + minutes) + seconds

        file_start_adjustment = (((first_double_eight + 1) * 10 + first_eight) * (fs / 1000) + index) / fs

        tpm -= file_start_adjustment

        return tpm, julian_date

    # ---------------------------- Protected functions for feature calculation -----------------------------------------

    def _calculate_signal_envelope(self):
        #   Calculate the energy envelope of the signal that is required for many of the features

        analytic_signal = scipy.signal.hilbert(self.samples)
        amplitude_modulation = np.abs(analytic_signal)
        normalized_freq = self.cutoff_frequency / (self.sample_rate / 2)
        sos = scipy.signal.butter(3, normalized_freq, btype='low', analog=False, output='sos')
        self._signal_envelope = scipy.signal.sosfilt(sos, amplitude_modulation)

        #   Normalize the envelope

        self._normal_signal_envelope = self.signal_envelope / np.max(self.signal_envelope)

    def _trim_by_samples(self, s0: int = None, s1: int = None):
        """
        This function will trim the waveform and return a subset of the current waveform based on sample indices within
        the 'samples' property within this class.

        Parameters
        __________
        :param s0: int - the start sample of the trimming. If s0 is None, then interface will use the first sample
        :param s1: int - the stop sample of the trimming. If s1 is None, then the interface uses the last sample

        Returns
        _______
        :returns: Waveform - a subset of the waveform samples
        """

        #   Handle the start/stop samples may be passed as None arguments

        if s0 is None:
            s0 = 0

        if s1 is None:
            s1 = self._samples.shape[0]

        #   Determine the new start time of the waveform

        if isinstance(self.start_time, datetime):
            t0 = self.start_time + timedelta(seconds=s0 / self.sample_rate)
        else:
            t0 = self.start_time + s0 / self.sample_rate

        #   Create the waveform based on the new time, and the subset of the samples

        return Waveform(self.samples[np.arange(s0, s1)].copy(), self.sample_rate, t0, remove_dc_offset=False)

    def _scale_waveform(self, scale_factor: float = 1.0, inplace: bool = False):
        """
        This function applies a scaling factor to the waveform's sample in a linear scale factor.

        Parameters
        __________
        :param scale_factor: float - the linear unit scale factor to change the amplitude of the sample values
        :param inplace: boolean - Whether to modify the samples within the current object, or return a new object

        Returns
        _______
        :returns: If inplace == True a new Waveform object with the sample magnitudes scaled, None otherwise
        """

        if inplace:
            self._samples *= scale_factor

            return None
        else:
            return Waveform(self._samples * scale_factor, self.sample_rate, self.start_time)

    # -------------------- Public functions for operations on the samples within the Waveform --------------------------

    def apply_calibration(self, wfm, level: float = 114, frequency: float = 1000, inplace: bool = False):
        return None

    def scale_signal(self, factor: float = 1.0, inplace: bool = False,
                     scale_type: scaling_method = scaling_method.linear):
        """
        This method will call the sub-function to scale the values of the waveform in linear fashion. If the scale
        factor is provided in logarithmic form, it will be converted to a linear value and sent to the sub-function.

        Parameters
        ----------
        :param factor: float - the scale factor that needs to be passed to the scaling sub-function
        :param inplace: bool - whether to manipulate the data within the current class, or return a new instance
        :param scale_type: scaling_method - how to apply the scaling to the signal

        Returns
        -------

        :returns: output of sub-function
        """

        scale_factor = factor

        if scale_type == scaling_method.logarithmic:
            scale_factor = 10**(scale_factor / 20)

        return self._scale_waveform(scale_factor, inplace)

    def trim(self, s0: float = 0.0, s1: float = None, method: trimming_methods = trimming_methods.samples):
        """
        This function will remove the samples before s0 and after s1 and adjust the start time
        :param s0: float - the sample index or time of the new beginning of the waveform
        :param s1: float - the sample index or time of the end of the new waveform
        :param method: trimming_methods - the method to trim the waveform
        :return: generic_time_waveform object
        """

        #   Determine whether to use the time or sample methods

        if method == trimming_methods.samples:
            return self._trim_by_samples(int(s0), int(s1))
        elif method == trimming_methods.times:
            t0 = s0
            t1 = s1

            if isinstance(self.start_time, datetime):
                start_seconds = 60 * (60 * self.start_time.hour + self.start_time.minute) + self.start_time.second + \
                                self.start_time.microsecond / 1e6
            else:
                start_seconds = self.start_time

            s0 = (t0 - start_seconds) * self.sample_rate
            ds = (t1 - t0) * self.sample_rate
            s1 = s0 + ds

            return self._trim_by_samples(int(s0), int(s1))

    def apply_window(self, window: windowing_methods = windowing_methods.hanning, windowing_parameter=None):
        """
        This will apply a window with the specific method that is supplied by the window argument and returns a
        generic_time_waveform with the window applied

        :param window:windowing_methods - the enumeration that identifies what type of window to apply to the waveform
        :param windowing_parameter: int or float - an additional parameter that is required for the window
        :returns: generic_time_waveform - the waveform with the window applied
        """

        W = []

        if window == windowing_methods.tukey:
            W = tukey(len(self.samples), windowing_parameter)

        elif window == windowing_methods.rectangular:
            W = tukey(len(self.samples), 0)

        elif window == windowing_methods.hanning:
            W = tukey(len(self.samples), 1)

        elif window == windowing_methods.hamming:
            W = hamming(len(self.samples))

        return Waveform(self.samples * W, self.fs, self.start_time)

    def apply_iir_filter(self, b, a):
        """
        This function will be able to apply a filter to the samples within the file and return a new
        generic_time_waveform object

        :param b: double, array-like - the forward coefficients of the filter definition
        :param a: double, array-like - the reverse coefficients of the filter definition
        """

        self._forward_coefficients = b
        self._reverse_coefficients = a
        return Waveform(scipy.signal.lfilter(b, a, self.samples), self.sample_rate, self.start_time)

    def apply_a_weight(self):
        """
        This function specifically applies the a-weighting filter to the acoustic data, and returns a new waveform with
        the filter applied.

        :returns: generic_time_waveform - the filtered waveform
        """
        a, c = Waveform.AC_Filter_Design(self.sample_rate)

        return self.apply_iir_filter(a[0], a[1])

    def apply_c_weight(self):
        """
        This function specifically applies the a-weighting filter to the acoustic data, and returns a new waveform with
        the filter applied.

        :returns: generic_time_waveform - the filtered waveform
        """
        a, c = Waveform.AC_Filter_Design(self.sample_rate)

        return self.apply_iir_filter(c[0], c[1])

    def apply_lowpass(self, cutoff: float, order: int = 4):
        """
        This function applies a Butterworth filter to the samples within this class.

        :param cutoff: double - the true frequency in Hz
        :param order: double (default: 4) - the order of the filter that will be created and applied

        :returns: generic_time_waveform - the filtered waveform
        """

        #   Determine the nyquist frequency

        nyquist = self.sample_rate / 2.0

        #   Determine the normalized frequency

        normalized_cutoff = cutoff / nyquist

        #   Design the filter

        b, a = scipy.signal.butter(order, normalized_cutoff, btype='low', analog=False, output='ba')

        #   Filter the data and return the new waveform object

        return self.apply_iir_filter(b, a)

    def apply_highpass(self, cutoff: float, order: int = 4):
        """
        This function applies a Butterworth filter to the samples within this class.

        :param cutoff: double - the true frequency in Hz
        :param order: double (default: 4) - the order of the filter that will be created and applied

        :returns: Waveform - the filtered waveform
        """

        #   Determine the nyquist frequency

        nyquist = self.sample_rate / 2.0

        #   Determine the normalized frequency

        normalized_cutoff = cutoff / nyquist

        #   Design the filter

        b, a = scipy.signal.butter(order, normalized_cutoff, btype='high', analog=False, output='ba')

        #   Filter the data and return the new waveform object

        return self.apply_iir_filter(b, a)

    def apply_bandpass(self, low_cutoff: float, high_cutoff: float, order: int = 3):
        """
        This function determines the bandpass Butterworth filter coefficients and sends them and the current waveform
        into the function that will filter the data with an IIR filter coefficient set

        Parameters
        ----------
        low_cutoff: float - the regular frequency cutoff for the low edge of the band pass filter (Units: Hz)
        high_cutoff: float - the regular frequency cutoff for the upper edge of the band pass filter (Units: Hz)
        order: int - default: 3, the order of the filter
        """

        #   Determine the nyquist frequency for the upper and lower edges of the band
        nyquist = self.sample_rate / 2.0
        upper = high_cutoff / nyquist
        lower = low_cutoff / nyquist

        #   Design the filter
        b, a = scipy.signal.butter(order, [lower, upper], btype='bandpass', analog=False, output='ba')

        #   send this waveform and the coefficients into the filtering algorithm and return the filtered waveform
        return self.apply_iir_filter(b, a)

    def resample(self, new_sample_rate: int):
        """
        This function resamples the waveform and returns a new object with the correct sample rate and sample count.
        This function employs the resample function within scipy.signal to conduct the resampling.

        Parameters
        ----------
        new_sample_rate: int - the new sample rate that we want to create a signal for

        Returns
        -------
        Waveform - the new waveform object that contains the resampled data with the new sample rate.
        """

        #   Determine the ratio of the current sample rate to the new sample rate

        sr_ratio = new_sample_rate / self.sample_rate

        return Waveform(scipy.signal.resample(self.samples, int(np.floor(len(self.samples) * sr_ratio))),
                        new_sample_rate, self.start_time)

    def apply_tob_equalizer_filter(self, frequencies: np.ndarray, sound_pressure_levels: np.ndarray):
        from .spectral.fractional_octave_band_tools import FractionalOctaveBandTools as fob
        import warnings

        """
        Pass the waveform through a series of filters adjusting the amplitude by the adjusted value of the arrays.

        Parameters
        ----------
        frequencies: np.ndarray - the collection of frequencies to generate filters
        sound_pressure_levels: np.ndarray - the collection of sound pressure levels to adjust the amplitude

        Returns
        -------
        Waveform - the summation of the n Waveforms that are generated by each of the filters.
        """
        #   Adjust the sound pressure level array by the average
        magnitude = sound_pressure_levels - np.mean(sound_pressure_levels)

        #   Determine the center frequency of the highest octave containing the upper frequency
        full_band = int(np.floor(fob.nearest_band(1, frequencies[-1])))

        #   Now determine the upper edge of this full octave band
        f_hi = fob.upper_frequency(1, full_band)
        f_lo = fob.lower_frequency(1, full_band)

        #   With this information we can determine the band number of the highest one-third-octave band in this full
        #   octave band. This is used to determine the filter collection.
        b_coefficients = list()
        a_coefficients = list()

        f_band = int(np.floor(fob.nearest_band(3, f_hi)))

        #   Create the collection of samples that we will add the output of the filter to create the final object
        samples = np.zeros((len(self.samples),))

        #   Now loop through the bands until we exit this full octave band
        while fob.lower_frequency(3, f_band) >= f_lo * 0.9:
            #   Define the window for the bandpass filter
            upper = fob.upper_frequency(3, f_band)
            lower = fob.lower_frequency(3, f_band)
            window = np.array([lower, upper]) / self.sample_rate / 2.0

            #   Create the filter coefficients for this frequency band and add it to the list for each coefficient set
            b, a = scipy.signal.butter(
                3,
                window,
                btype='bandpass',
                analog=False,
                output='ba'
            )

            b_coefficients.append(b)
            a_coefficients.append(a)

            #   Decrement the band number to move to the next band down.
            f_band -= 1

            #   With that we now need to determine the limits of the band calculations
            #   Determine the octave bands that will need to be calculated to cover the desired frequency range.
            low_band = int(np.floor(fob.nearest_band(1, frequencies[0])))
            hi_band = int(np.floor(fob.nearest_band(1, frequencies[-1])))

            #   Get the index of the band at the top of the full octave filter
            fob_band_index = int(np.floor(fob.nearest_band(3, fob.upper_frequency(1, hi_band))))

            #   Make a copy of the waveform that can be decimated
            wfm = Waveform(pressures=self.samples.copy(),
                           sample_rate=self.sample_rate,
                           start_time=self.start_time)

            #   Loop through the frequencies
            for band_index in range(hi_band, low_band - 1, -1):
                #   Loop through the filter definitions
                for filter_index in range(len(b_coefficients)):
                    filtered_waveform = wfm.apply_iir_filter(b_coefficients[filter_index], a_coefficients[filter_index])

                    band_index -= 1

                    #   Up sample the waveform
                    while filtered_waveform.sample_rate < self.sample_rate:
                        filtered_waveform = filtered_waveform.resample(filtered_waveform.sample_rate * 2)

                    #   Apply the amplification factor to the data
                    filtered_waveform.scale_signal(magnitude[band_index], scale_type=scaling_method.logarithmic)

                    #   Add the contents of the waveform to the sample list
                    if (fob.center_frequency(3, band_index) <= frequencies[-1]) and \
                            (fob.center_frequency(3, band_index) >= frequencies[0]):
                        samples += filtered_waveform.samples[:len(samples)]

                    #   Decimate the waveform, halving the sample rate and making the filter definitions move down a
                    #   full octave
                    if len(wfm.samples) / 2 < 3 * len(b_coefficients):
                        warnings.warn(
                            "The number of points within the Waveform are insufficient to calculate digital filters "
                            "lower than these frequencies")

                        break

                    wfm = Waveform(pressures=scipy.signal.decimate(wfm.samples, 2),
                                   sample_rate=wfm.sample_rate,
                                   start_time=wfm.start_time)

        return Waveform(pressures=samples, sample_rate=self.sample_rate, start_time=self.start_time)

    def is_calibration(self):
        """
        This function examines the samples and determines whether the single contains a single pure tone.  If it does
        the function returns the approximate frequency of the tone.  This will examine every channel and determine
        whether each channel is a calibration tone

        :returns: bool - flag determining whether the signal was pure tone
                  float - the approximate frequency of the pure tone
        """

        calibration = None
        freq = None

        #   Loop through the channels

        #   To remove high frequency transients, we pass the signal through a 2 kHz low pass filter

        wfm = Waveform(self.samples, self.sample_rate, self.start_time)
        wfm.apply_lowpass(2000)

        peaks = scipy.signal.find_peaks(wfm.samples, height=0.8 * np.max(self.samples))[0]

        if len(peaks) >= 2:
            calibration = False
            freq = -1

            #   Determine the distance between any two adjacent peaks

            distance_sample = np.diff(peaks)

            #   Determine the distance between the samples in time

            distance_time = distance_sample / self.sample_rate

            #   Determine the frequencies

            frequencies = 1 / distance_time

            freq = np.mean(frequencies)

            calibration = (abs(freq - 1000) < 0.1 * 1000) or \
                          (abs(freq - 250) < 0.1 * 250)

        return calibration, freq

    def get_features(self, include_sq_metrics: bool = True):
        """
        This function calculates the various features within the global time analysis and stores the results in the
        class object.  At the end, a dictionary of the values is available and returned to the calling function.

        Returns
        -------
        features : dict()
            The dictionary containing the various values calculated within this method.
        """

        #   Create the dictionary that will hold the data for return to the user

        features = {'attack': self.attack,
                    'decrease': self.decrease,
                    'release': self.release,
                    'log_attack': self.log_attack,
                    'attack slope': self.attack_slope,
                    'decrease slope': self.decrease_slope,
                    'temporal centroid': self.temporal_centroid,
                    'effective duration': self.effective_duration,
                    'amplitude modulation': self.amplitude_modulation,
                    'frequency modulation': self.frequency_modulation,
                    'auto-correlation': self.auto_correlation,
                    'zero crossing rate': self.zero_crossing_rate}

        if include_sq_metrics:
            features['loudness'] = self.loudness
            features['roughness'] = self.roughness
            features['sharpness'] = self.sharpness

        return features

    def calculate_temporal_centroid(self):

        env_max_idx = np.argmax(self.signal_envelope)
        over_threshold_idcs = np.where(self.normal_signal_envelope > self.centroid_threshold)[0]

        over_threshold_start_idx = over_threshold_idcs[0]
        if over_threshold_start_idx == env_max_idx:
            over_threshold_start_idx = over_threshold_start_idx - 1

        over_threshold_end_idx = over_threshold_idcs[-1]

        over_threshold_TEE = self.signal_envelope[over_threshold_start_idx - 1:over_threshold_end_idx - 1]
        over_threshold_support = [*range(len(over_threshold_TEE))]
        over_threshold_mean = np.divide(np.sum(np.multiply(over_threshold_support, over_threshold_TEE)),
                                        np.sum(over_threshold_TEE))

        temporal_threshold = ((over_threshold_start_idx + 1 + over_threshold_mean) / self.sample_rate)

        return temporal_threshold

    def calculate_effective_duration(self):

        env_max_idx = np.argmax(self.signal_envelope)
        over_threshold_idcs = np.where(self.normal_signal_envelope > self.effective_duration_threshold)[0]

        over_threshold_start_idx = over_threshold_idcs[0]
        if over_threshold_start_idx == env_max_idx:
            over_threshold_start_idx = over_threshold_start_idx - 1

        over_threshold_end_idx = over_threshold_idcs[-1]

        return (over_threshold_end_idx - over_threshold_start_idx + 1) / self.sample_rate

    def rms_envelope(self):

        win_size = int(round(self._window_size_seconds * self.sample_rate))
        hop_size = int(round(self._hop_size_seconds * self.sample_rate))
        t_support = [*range(0, int(round(hop_size * np.floor((len(self.samples) - win_size) / hop_size))), hop_size)]
        value = []
        for i in range(len(t_support)):
            a = self.samples[np.subtract(np.add(t_support[i], [*range(1, win_size + 1)]), 1).astype(int)]
            value.append(np.sqrt(np.mean(np.power(a, 2))))
        t_support = np.divide(np.add(t_support, np.ceil(np.divide(win_size, 2))),
                              self.sample_rate)
        return value

    def instantaneous_temporal_features(self):
        """
        This function will calculate the instantaneous features within the temporal analysis.  This includes the
        auto-correlation and the zero crossing rate.
        """
        count = 0
        dAS_f_SupX_v_count = 0
        temporal_feature_times = np.zeros(
            (int(np.floor((len(self.samples) - self.window_size_samples) / self.hop_size_samples) + 1),))

        auto_coefficients = np.zeros((len(temporal_feature_times), self.coefficient_count))
        zero_crossing_rate = np.zeros((len(temporal_feature_times),))

        #   Loop through the frames

        for n in range(0, len(temporal_feature_times)):
            #   Get the frame

            frame_length = self.window_size_samples
            start = n * self.hop_size_samples
            frame_index = np.arange(start, frame_length + start)
            f_Frm_v = self.samples[frame_index] * np.hamming(self.window_size_samples)
            temporal_feature_times[n] = n * self.hop_size_seconds

            count += 1

            #   Calculate the auto correlation coefficients

            auto_coefficients[n, :] = sm.tsa.acf(f_Frm_v, nlags=self.coefficient_count, fft=False)[1:]

            #   Now the zero crossing rate

            i_Sign_v = np.sign(f_Frm_v - np.mean(f_Frm_v))
            i_Zcr_v = np.where(np.diff(i_Sign_v))[0]
            i_Num_Zcr = len(i_Zcr_v)
            zero_crossing_rate[n] = i_Num_Zcr / (len(f_Frm_v) / self.sample_rate)

        return temporal_feature_times, auto_coefficients, zero_crossing_rate

    def calculate_modulation(self):
        """
        Calculate the frequency/amplitude modulations of the signal.  This can be accomplished with either a Fourier or
        Hilbert method.

        Returns
        -------

        frequency_modulation : double
            A metric measuring the frequency modulation of the signal
        amplitude_modulation : double
            A metric measuring the amplitude modulation of the signal
        """

        sample_times = np.arange(len(self.signal_envelope) - 1) / self.sample_rate

        sustain_start_time = self._Addresses[1]
        sustain_end_time = self._Addresses[4]

        is_sustained = False

        if (sustain_end_time - sustain_start_time) > 0.02:
            pos_v = np.where((sustain_start_time <= sample_times) & (sample_times <= sustain_end_time))[0]
            if len(pos_v) > 0:
                is_sustained = True

        if not is_sustained:
            amplitude_modulation = 0
            frequency_modulation = 0
        else:
            envelop_v = self.signal_envelope[pos_v]
            temps_sec_v = sample_times[pos_v]
            M = np.mean(envelop_v)

            #   Taking the envelope

            mon_poly = np.polyfit(temps_sec_v, np.log(envelop_v), 1)
            hat_envelope_v = np.exp(np.polyval(mon_poly, temps_sec_v))
            signal_v = envelop_v - hat_envelope_v

            sa_v = scipy.signal.hilbert(signal_v)
            sa_amplitude_v = abs(signal_v)
            sa_phase_v = np.unwrap(np.angle(sa_v))
            sa_instantaneous_frequency = (1 / 2 / np.pi) * sa_phase_v / (len(temps_sec_v) / self.sample_rate)

            amplitude_modulation = np.median(sa_amplitude_v)
            frequency_modulation = np.median(sa_instantaneous_frequency)

        return frequency_modulation, amplitude_modulation

    def calculate_log_attack(self):
        """
        This calculates the various global attributes.

        In some cases the calculation of the attack did not return an array, so
        the error is trapped for when a single values is returned rather than
        an array.

        Returns
        -------
        attack_start : TYPE
            DESCRIPTION.
        log_attack_time : TYPE
            DESCRIPTION.
        attack_slope : TYPE
            DESCRIPTION.
        attack_end : TYPE
            DESCRIPTION.
        release : TYPE
            DESCRIPTION.
        release_slope : TYPE
            DESCRIPTION.

        """

        #   Define some specific constants for this calculation

        method = 3
        noise_threshold = 0.15
        decrease_threshold = 0.4

        #   Calculate the position for each threshold

        percent_step = 0.1
        percent_value_value = np.arange(percent_step, 1 + percent_step, percent_step)
        percent_value_position = np.zeros(percent_value_value.shape)

        for p in range(len(percent_value_value)):
            percent_value_position[p] = np.where(self.normal_signal_envelope >= percent_value_value[p])[0][0]

        #   Detection of the start (start_attack_position) and stop (end_attack_position) of the attack

        position_value = np.where(self.normal_signal_envelope > noise_threshold)[0]

        #   Determine the start and stop positions based on selected method

        if method == 1:  # Equivalent to a value of 80%
            start_attack_position = position_value[0]
            end_attack_position = position_value[int(np.floor(0.8 / percent_step))]
        elif method == 2:  # Equivalent to a value of 100%
            start_attack_position = position_value[0]
            end_attack_position = position_value[int(np.floor(1.0 / percent_step))]
        elif method == 3:
            #   Define parameters for the calculation of the search for the start and stop of the attack

            # The terminations for the mean calculation

            m1 = int(round(0.3 / percent_step)) - 1
            m2 = int(round(0.6 / percent_step))

            #   define the multiplicative factor for the effort

            multiplier = 3

            #   Terminations for the start attack correction

            s1att = int(round(0.1 / percent_step)) - 1
            s2att = int(round(0.3 / percent_step))

            #   Terminations for the end attack correction

            e1att = int(round(0.5 / percent_step)) - 1
            e2att = int(round(0.9 / percent_step))

            #   Calculate the effort as the effective difference in adjacent position values

            dpercent_position_value = np.diff(percent_value_position)

            #   Determine the average effort

            M = np.mean(dpercent_position_value[m1:m2])

            #   Start the start attack calculation
            #   we START JUST AFTER THE EFFORT TO BE MADE (temporal gap between percent) is too large

            position2_value = np.where(dpercent_position_value[s1att:s2att] > multiplier * M)[0]

            if len(position2_value) > 0:
                index = int(np.floor(position2_value[-1] + s1att))
            else:
                index = int(np.floor(s1att))

            start_attack_position = percent_value_position[index]

            #   refinement: we are looking for the local minimum

            delta = int(np.round(0.25 * (percent_value_position[index + 1] - percent_value_position[index]))) - 1
            n = int(np.floor(percent_value_position[index]))

            if n - delta >= 0:
                min_position = np.argmin(self.normal_signal_envelope[n - delta:n + delta])
                start_attack_position = min_position + n - delta - 1

            #   Start the end attack calculation
            #   we STOP JUST BEFORE the effort to be made (temporal gap between percent) is too large

            position2_value = np.where(dpercent_position_value[e1att:e2att] > multiplier * M)[0]

            if len(position2_value) > 0:
                index = int(np.floor(position2_value[0] + e1att))
            else:
                index = int(np.floor(e1att))

            end_attack_position = percent_value_position[index]

            #   refinement: we are looking for the local minimum

            delta = int(np.round(0.25 * (percent_value_position[index] - percent_value_position[index - 1])))
            n = int(np.floor(percent_value_position[index]))

            if n - delta >= 0:
                min_position = np.argmax(self.normal_signal_envelope[n - delta:n + delta + 1])
                end_attack_position = min_position + n - delta

        #   Calculate the Log-attack time

        if start_attack_position == end_attack_position:
            start_attack_position -= 1

        rise_time_n = end_attack_position - start_attack_position
        log_attack_time = np.log10(rise_time_n / self.sample_rate)

        #   Calculate the temporal growth - New 13 Jan 2003
        #   weighted average (Gaussian centered on percent=50%) slopes between satt_posn and eattpos_n

        start_attack_position = int(np.round(start_attack_position))
        end_attack_position = int(np.round(end_attack_position))

        start_attack_value = self.normal_signal_envelope[start_attack_position]
        end_attack_value = self.normal_signal_envelope[end_attack_position]

        threshold_value = np.arange(start_attack_value, end_attack_value, 0.1)
        threshold_position_seconds = np.zeros(np.size(threshold_value))

        for i in range(len(threshold_value)):
            position = \
                np.where(self.normal_signal_envelope[start_attack_position:end_attack_position] >= threshold_value[i])[
                    0][0]
            threshold_position_seconds[i] = position / self.sample_rate

        slopes = np.divide(np.diff(threshold_value), np.diff(threshold_position_seconds))

        #   Calculate the increase

        thresholds = (threshold_value[:-1] + threshold_value[1:]) / 2
        weights = np.exp(-(thresholds - 0.5) ** 2 / (0.5 ** 2))
        increase = np.sum(np.dot(slopes, weights)) / np.sum(weights)

        #   Calculate the time decay

        envelope_max_index = np.where(self.normal_signal_envelope == np.max(self.normal_signal_envelope))[0]
        envelope_max_index = int(np.round(0.5 * (envelope_max_index + end_attack_position)))

        stop_position = np.where(self.normal_signal_envelope > decrease_threshold)[0][-1]

        if envelope_max_index == stop_position:
            if stop_position < len(self.normal_signal_envelope):
                stop_position += 1
            elif envelope_max_index > 1:
                envelope_max_index -= 1

        #   Calculate the decrease

        X = np.arange(envelope_max_index, stop_position + 1) / self.sample_rate
        X_index = np.arange(envelope_max_index, stop_position + 1)
        Y = np.log(self.normal_signal_envelope[X_index])
        polynomial_fit = np.polyfit(X, Y, 1)
        decrease = polynomial_fit[0]

        #   Create the list of addresses that we are interested in storing for later consumption

        addresses = np.array([start_attack_position, envelope_max_index, 0, 0, stop_position]) / self.sample_rate

        return log_attack_time, increase, decrease, addresses

    def overall_level(self, integration_time: float = None,
                      weighting: weighting_function = weighting_function.unweighted):
        """
        Integrate the levels within the waveform to generate the weighted level. This will permit different weighting
        functions to be applied before the calculation of the overall level.

        Parameters
        ----------
        integration_time: float, default: None - The amount of time that we will collect prior to determining the
            RMS level within the samples.
        weighting: weighting_function, default: unweighted - the weighting to be applied prior to determining the RMS
            value of the signal.

        Returns
        -------
        float, array-like - A collection of the overall levels, with applicable weighting, with the number being equal
            to the int(np.floor(duration / integration_time))

        Revision
        20221007 - FSM - updated the method when the start_time is a datetime rather than a floating point value
        """

        if integration_time is None:
            n = 1
            integration_time = self.duration
        else:
            n = int(np.floor(self.duration / integration_time))

        if weighting == weighting_function.a_weighted:
            wfm = self.apply_a_weight()
        elif weighting == weighting_function.c_weighted:
            wfm = self.apply_c_weight()
        else:
            wfm = Waveform(self.samples, self.sample_rate, self.start_time)
        level = list()

        t0 = self.start_time
        if isinstance(t0, datetime):
            t0 = 60 * (60 * t0.hour + t0.minute) + t0.second + t0.microsecond / 1e6

        for i in range(n):
            subset = wfm.trim(t0, t0 + integration_time, trimming_methods.times)
            level.append(np.std(subset.samples))

            t0 += integration_time

        return 20 * np.log10(np.array(level) / 20e-6)

    def cross_correlation(self, b, mode=correlation_mode.valid, lag_limit=None):
        """
        This function determines the cross correlation between the current waveform and the waveform passed to the
        function.

        Parameters
        ----------
        b: Waveform - the signal to compare to the current waveform's samples
        mode: correlation_mode - the mode of the correlation that we want to execute for the correlation methods
        lag_limit: - the limit of the correlation analysis

        Returns
        -------

        value of the maximum correlation
        sample lag of the maximum correlation

        Remarks
        2022-12-01 - FSM - Added completed enumeration usage for different correlation modes
        """

        # TODO - @Alan - we need a test for this function.

        if not isinstance(b, Waveform):
            raise ValueError("The first argument is required to be a Waveform object")

        sig = b.samples
        ref_sig = self.samples
        if len(sig) > len(ref_sig):
            sig, ref_sig = ref_sig, sig

        M = len(ref_sig)
        N = len(sig)

        if lag_limit is None:
            correlation_values = np.correlate(ref_sig, sig, mode.name)
            if mode == correlation_mode.valid:
                lags = np.arange(0, max(M, N) - min(M, N) + 1)
            elif mode == correlation_mode.full:
                lags = np.arange(-(N - 1), M)
            elif mode == correlation_mode.same:
                lags = np.arange(-np.floor(N / 2), M - np.floor(N / 2))
        else:
            ref_sig_pad = np.pad(ref_sig.conj(), lag_limit, mode='constant')
            correlation_values = np.zeros(2 * lag_limit + 1)
            for i in range(0, 2 * lag_limit + 1):
                correlation_values[i] = sum(ref_sig_pad[i:len(sig) + i] * sig)
            lags = np.arange(-lag_limit, lag_limit + 1)

        return np.max(correlation_values), lags[np.argmax(correlation_values)]

    #   ----------------------------------------------- Operators ------------------------------------------------------

    def __add__(self, other):
        """
        This function will add the contents of one waveform to the other. This feature checks the sample rate to ensure
        that they both possess the same sample times. Also, if the data starts at different times, this function will
        create a new object that is the addition of the samples, with the new sample times.

        Parameters
        ----------
        :param other: Waveform - the new object to add to this class's data

        Returns
        -------
        :returns: - A new Waveform object that is the sum of the two
        """

        if not isinstance(other, Waveform):
            ValueError("You must provide a new Waveform object to add to this object.")

        if self.sample_rate != other.sample_rate:
            ValueError("At this time, the two waveforms must possess the same sample rate to add them together")

        s0 = int(other.start_time * other.sample_rate)
        s1 = s0 + len(other.samples)

        return Waveform(self.samples[s0:s1] + other.samples, self.sample_rate, self.start_time)

    def __sub__(self, other):
        """
        This function subtracts another Waveform object from the current object and returns the value as a new
        Waveform. If the start times, durations, or sample rates are not equal then the function returns a ValueError

        """

        if self.start_time != other.start_time or self.duration != other.duration or self.sample_rate != \
                other.sample_rate:
            raise ValueError("The meta-data of these two waveforms is inconsistent making it impossible to know how "
                             "to subtract the information in the pressures.")

        return Waveform(self.samples - other.samples, self.sample_rate, self.start_time, False)

    def resample(self, new_sample_rate: int):
        """
        This function resamples the waveform and returns a new object with the correct sample rate and sample count.
        This function employs the resample function within scipy.signal to conduct the resampling.

        Parameters
        ----------
        new_sample_rate: int - the new sample rate that we want to create a signal for

        Returns
        -------
        Waveform - the new waveform object that contains the resampled data with the new sample rate.
        """

        #   Determine the ratio of the current sample rate to the new sample rate

        sr_ratio = new_sample_rate / self.sample_rate

        return Waveform(scipy.signal.resample(self.samples, int(np.floor(len(self.samples) * sr_ratio))),
                        new_sample_rate, self.start_time)




import numpy
import numpy as np
import datetime


class AcousticWeights:
    """
    This class contains a number of calculations for various community noise metrics that are useful for a variety of
    calculations across the study of acoustics.

    FSM - the math module was replaced with references to the numpy module to facilitate use with numpy arrays and the
    pandas DataFrame objects
    """

    #TODO: Frank - make this more generic
    @staticmethod
    def sound_exposure_level(times, levels, dB_down):
        """
        The sound exposure level attempts to determine the equivalent level of the acoustic energy placed within a
        single second of the acoustic level.  The dB_down parameter determines how far below the peak that the algorithm
        seeks to integrate the data.

        times : datetime, array-like
            a collection of datetime objects that represent the times for the acoustic levels
        levels : double, array-like
            a collection of acoustic levels that are selected at the same time values as the times array
        dB_down : double
            the number of decibels below the peak that we will integrate the acoustics levels

        returns : double
            the integrated level between the times marking the location of the dB_down levels.
        """

        #   Find the indices for the integration

        start_index, stop_index = AcousticWeights.find_dB_down_limits(levels, dB_down)

        #   Determine the equivalent level between these times
        if isinstance(times[0], datetime.datetime):
            tin = (times[stop_index] - times[start_index]).total_seconds()
        else:
            tin = times[stop_index] - times[start_index]

        return AcousticWeights.leq(
            levels,
            tin,
            1,
            start_index,
            stop_index)

    @staticmethod
    def find_dB_down_limits(levels, dB_down_level):
        """
        Examine the array of levels and determine the points that were above the peak - dB_down_level

        levels : double, array-like
            the acoustic levels in an array that will be examined
        dB_down_level : double
            the level below the peak that will set the limits of the integration

        returns: double, tuple
            the start and stop index of the points to integrate
        """

        #   Find the maximum level

        max_level = max(levels)

        #   Find the index of the maximum value

        max_index = np.argmax(levels)

        #   Determine the start_index

        start_index = -1
        for i in range(max_index, -1, -1):
            if levels[i] <= (max_level - dB_down_level):
                start_index = i
                break

        #   Determine the stop_index

        stop_index = -1
        for i in range(max_index, len(levels), 1):
            if levels[i] <= (max_level - dB_down_level):
                stop_index = i
                break

        #   Apply some constraints to ensure that we are within the limits of the array

        if start_index < 0:
            start_index = 0
        if stop_index < 0:
            stop_index = len(levels) - 1

        #   Return the arrays

        return start_index, stop_index

    @staticmethod
    def leq(levels, tin, tout, start_index, stop_index):
        """
        The equivalent level is an integration of levels changing the temporal resolution of the acoustic levels.

        levels : double, array-like
            the list of acoustic levels
        tin : double
            the temporal integration of the input level
        tout : double
            the resultant temporal integration of the output level
        start_index : int
            the index within the levels array that we will begin the integration
        stop_index : int
            the index within the levels array that we will stop the integration

        returns : double
            the integrated, equivalent level
        """

        #   Initialize the acoustic equivalent level

        leq = 0.0

        #   Sum the linear elements units of sound

        for i in range(start_index, stop_index + 1, 1):
            leq += 10.0**(levels[i] / 10.0)

        #   apply the logarithmic conversion and the application of the temporal ratio

        return 10 * numpy.log10(leq) + 10 * numpy.log10(tin / tout)

    @staticmethod
    def lf(spl):
        """
        Compute the equal weighted acoustic level across the level array

        spl : double, array-like
            the sound pressure levels across a specific frequency array.  The frequencies are not provided for this
            function because there is an equal weighting across all frequencies.

        return : double
            the overall acoustic level with equal weighting
        """

        x = np.asarray(spl).copy()

        x /= 10.0
        x = 10.0 ** x
        x = np.sum(x, axis=len(x.shape) - 1)
        return 10 * np.log10(x)

    @staticmethod
    def la(spl, frequency):
        """
        Compute the A-weighted acoustic level across the level array

        spl : double, array-like
            the sound pressure levels across a specific frequency array
        frequency : double, array-like
            the array of frequencies to calculate the weighting

        return : double
            the overall acoustic level with equal weighting
        """

        x = np.asarray(spl).copy()
        if len(x.shape) == 1:
            x = np.reshape(x, (len(x), 1)).transpose()

        weights = np.ones((x.shape[0], 1)).dot(np.reshape(AcousticWeights.aw(frequency),
                                                          (len(frequency), 1)).transpose())

        x += weights
        x /= 10
        x = 10.0 ** x
        x = np.sum(x, axis=(len(x.shape) - 1))
        return 10.0 * np.log10(x)

    @staticmethod
    def aw(frequency):
        """
        Given a frequency, determine the A-weighted correction for the acoustic level

        frequency : double, possible array-like
            the number of cycles per second to calculate the weight at
        """

        frequency = np.asarray(frequency).copy()

        f2 = 107.65265
        f3 = 737.86223
        K3 = 1.562339
        numerator = K3 * frequency ** 4.0
        denominator = (frequency ** 2 + f2 ** 2) * (frequency ** 2 + f3 ** 2)

        return 10 * numpy.log10(numerator / denominator) + AcousticWeights.cw(frequency)

    @staticmethod
    def cw(frequency):
        """
        Given a frequency, determine the C-weighted correction for the acoustic level

        frequency : double, possible array-like
            the number of cycles per second to calculate the weight at
        """
        f1 = 20.598997
        f4 = 12194.22
        K1 = 2.24e16
        numerator = K1 * frequency**4
        denominator = ((frequency**2 + f1**2)**2.0) * ((frequency**2 + f4**2)**2.0)

        return 10 * numpy.log10(numerator / denominator)

    @staticmethod
    def pnl(dSPL):
        """
        Determine the single number perceived noise level (PNL) based on the conversion from dB to Noys

        dSPL : double, array-like
            the sound pressure levels from 10 Hz to 10 kHz

        returns : double
            returns the perceived noise level in NOYS

        Remarks:
        2022-12-13 - FSM - Changed the end of the code to ensure that there is a non-infinite results when the sume of
            the levels is zero because the level is too quiet.
        """

        Ld = [49, 44, 39, 34, 30, 27, 24, 21, 18, 16, 16, 16, 16, 16, 15, 12, 9, 5, 4, 5, 6, 10, 17, 21]
        Le = [55, 51, 46, 42, 39, 36, 33, 30, 27, 25, 25, 25, 25, 25, 23, 21, 18, 15, 14, 14, 15, 17, 23, 29]
        Lb = [64, 60, 56, 53, 51, 48, 46, 44, 42, 40, 40, 40, 40, 40, 38, 34, 32, 30, 29, 29, 30, 31, 37, 41]
        La = [91.01, 85.88, 87.32, 79.85, 79.76, 75.96, 73.96, 74.91, 94.63, 1000, 1000, 1000, 1000, 1000, 1000, 1000,
              1000, 1000, 1000, 1000, 1000, 1000, 44.29, 50.72]
        Lc = [52, 51, 49, 47, 46, 45, 43, 42, 41, 40, 40, 40, 40, 40, 38, 34, 32, 30, 29, 29, 30, 31, 34, 37]
        Md = [0.079520, 0.068160, 0.068160, 0.059640, 0.053013, 0.053013, 0.053013, 0.053013, 0.053013, 0.053013,
              0.053013, 0.053013, 0.053013, 0.053013, 0.059640, 0.053013, 0.053013, 0.047712, 0.047712, 0.053013,
              0.053013, 0.068160, 0.079520, 0.059640]
        Me = [0.058098, 0.058098, 0.052288, 0.047534, 0.043573, 0.043573, 0.040221, 0.037349, 0.034859, 0.034859,
              0.034859, 0.034859, 0.034859, 0.034859, 0.034859, 0.040221, 0.037349, 0.034859, 0.034859, 0.034859,
              0.034859, 0.037349, 0.037349, 0.043573]
        Mc = [0.030103, 0.030103, 0.030103, 0.030103, 0.030103, 0.030103, 0.030103, 0.030103, 0.030103, 0, 0, 0, 0, 0,
              0, 0, 0, 0, 0, 0, 0, 0, 0.02996, 0.02996]
        Mb = [0.043478, 0.04057, 0.036831, 0.036831, 0.035336, 0.033333, 0.033333, 0.032051, 0.030675, 0.030103,
              0.030103, 0.030103, 0.030103, 0.030103, 0.030103, 0.02996, 0.02996, 0.02996, 0.02996, 0.02996, 0.02996,
              0.02996, 0.042285, 0.042285]
        dfn = [0.0 for i in range(31)]
        d_sum = 0.0
        fFJ = 0.15
        for j in range(0, 7, 1):
            dfn[j] = 0
        for i in range(7, 31, 1):
            if dSPL[i] > 250:
                return -1000

            if dSPL[i] >= La[i - 7]:
                dfn[i] = AcousticWeights.ILOG10(Mc[i - 7] * (dSPL[i] - Lc[i - 7]))

            elif Lb[i - 7] <= dSPL[i] < La[i - 7]:
                dfn[i] = AcousticWeights.ILOG10(Mb[i - 7] * (dSPL[i] - Lb[i - 7]))

            elif Le[i - 7] <= dSPL[i] < Lb[i - 7]:
                dfn[i] = AcousticWeights.ILOG10(Me[i - 7] * (dSPL[i] - Lb[i - 7]))

            elif Ld[i - 7] <= dSPL[i] < Le[i - 7]:
                dfn[i] = 0.1 * AcousticWeights.ILOG10(Md[i - 7] * (dSPL[i] - Ld[i - 7]))

            if abs(dfn[i]) > 300:
                dfn[i] = 0

            if abs(dfn[i]) < -300:
                dfn[i] = 0

            if abs(dfn[i]) < 1e-10:
                dfn[i] = 0

            if dfn[i] > 2048:
                return -2000

        for i in range(0, 31, 1):
            if dfn[i] == float('inf'):
                dfn[i] = 0.0
        damx = dfn[7]

        for i in range(7, 31, 1):
            if dfn[i] > damx:
                damx = dfn[i]
            d_sum += dfn[i]
        d_sum = (d_sum - damx) * fFJ + damx

        if d_sum == 0:
            d_sum = 1
        return 40 + 33.22 * numpy.log10(d_sum)

    @staticmethod
    def ILOG10(x):
        """
        Determine the inverse log10, or 10**x
        """

        return 10.0 ** x

    @staticmethod
    def tone_correction(dSPL):
        """
        This function determines the tone correction applied to the sound pressure level spectrum.  It is based on the
        description of the calculation within the FAR part 36, Appendix A36.4.3.1.

        dSPL : double, array-like
            the collection of sound pressure levels from 10 Hz to 10 kHz

        returns : double
            the single value tone correction for the spectrum to be applied to the integrated acoustic levels.
        """

        #   Step 1 - Calculate the changes in adjacent sound pressure levels (or slopes)

        slopes = [0.0 for x in range(len(dSPL))]
        for i in range(0, len(slopes), 1):
            if i < 10:
                slopes[i] = 0.0
            else:
                slopes[i] = dSPL[i] - dSPL[i - 1]

        #   Step 2 - find any slope changes greater than 5

        large_slope_changes = [False for x in range(len(dSPL))]
        for i in range(0, len(large_slope_changes), 1):
            if abs(slopes[i]) > 5:
                large_slope_changes[i] = True
            else:
                large_slope_changes[i] = False

        #   Step 3 - select the sound pressure level that needs to be corrected

        select_sound_pressure_level = [False for x in range(len(dSPL))]
        for i in range(1, len(select_sound_pressure_level), 1):
            if large_slope_changes[i]:
                if slopes[i] > 0 and slopes[i] > slopes[i - 1]:
                    select_sound_pressure_level[i] = True
                elif slopes[i] <= 0 and slopes[i - 1] > 0:
                    select_sound_pressure_level[i - 1] = True
                else:
                    select_sound_pressure_level[i] = False
            else:
                select_sound_pressure_level[i] = False

        #   Step 4 - Adjust selected sound pressure levels

        spl_prime = [0.0 for x in range(len(dSPL))]
        for i in range(0, len(select_sound_pressure_level), 1):
            if not select_sound_pressure_level[i]:
                spl_prime[i] = dSPL[i]
            else:
                if 8 < i < 30:
                    spl_prime[i] = 0.5 * (dSPL[i - 1] + dSPL[i + 1])
                elif i == 30:
                    spl_prime[i] = dSPL[i - 1] + slopes[i - 1]

        #   Step 5 - recompute new slopes

        slope_prime = [0.0 for x in range(len(dSPL) + 1)]
        for i in range(1, len(spl_prime), 1):
            slope_prime[i] = spl_prime[i] - spl_prime[i - 1]
        slope_prime[31] = slope_prime[30]

        #   Step 6 - compute the arithmetic mean of adjacent three slopes

        mean_slope = [0.0 for i in range(30)]
        for i in range(0, len(mean_slope), 1):
            mean_slope[i] = (1.0 / 3.0) * (slope_prime[i] + slope_prime[i + 1] + slope_prime[i + 2])

        # Step 7 - compute the final one-third-octave sound pressure levels

        final_spl = [0.0 for x in range(31)]
        for i in range(0, len(final_spl), 1):
            if i < 10:
                final_spl[i] = dSPL[i]
            else:
                final_spl[i] = final_spl[i - 1] + mean_slope[i - 1]

        #   Step 8 - calculate the differences between the original and final SPL values

        F = [0.0 for x in range(len(dSPL))]
        for i in range(0, len(F), 1):
            F[i] = dSPL[i] - final_spl[i]

        #   Step 9 - for each of the relevant one-third-octave bands, determine tone correction factors from the sound
        #   pressure level differences (F[i]) and the table in the FAR

        tone_corrections = [0.0 for x in range(len(dSPL))]
        for i in range(0, len(dSPL), 1):
            if i < 9:
                tone_corrections[i] = 0.0
            elif 9 <= i < 17:
                if F[i] < 1.5:
                    tone_corrections[i] = 0.0
                elif 1.5 <= F[i] < 3:
                    tone_corrections[i] = 0.5
                elif 3 <= F[i] < 20:
                    tone_corrections[i] = F[i] / 6.0
                elif 20 <= F[i]:
                    tone_corrections[i] = 3.0 + (1.0 / 3.0)
            elif 17 <= i <= 27:
                if F[i] < 1.5:
                    tone_corrections[i] = 0.0
                elif 1.5 <= F[i] < 3:
                    tone_corrections[i] = 1
                elif 3 <= F[i] < 20:
                    tone_corrections[i] = F[i] / 3
                elif 20 <= F[i]:
                    tone_corrections[i] = 2 * (3.0 + (1.0 / 3.0))
            elif i > 27:
                if F[i] < 1.5:
                    tone_corrections[i] = 0.0
                elif 1.5 <= F[i] < 3:
                    tone_corrections[i] = 0.5
                elif 3 <= F[i] < 20:
                    tone_corrections[i] = F[i] / 6.0
                elif 20 <= F[i]:
                    tone_corrections[i] = 3.0 + (1.0 / 3.0)

        #   Return the maximum of the corrections

        return max(tone_corrections)



import numpy as np
import struct
from ..waveform import Waveform
from .wavefile import WaveFile
import os.path
from datetime import datetime
from dateutil import parser


class StandardBinaryFile(Waveform):
    """
    This class holds the data for the standard ANSI Standard S12.75 formated
    binary files.  The class will permit access to the data, the header, and
    specific elements of the header that are required for various analyses that
    may be required.

    @author: frank mobley

    """

    def __init__(self, filename: str = None, wfm: Waveform = None, header_element_count: int = None,
                 sample_rate_key: str = 'SAMPLE RATE (HZ)', start_time_key: str = 'TIME (UTC ZULU)',
                 sample_format_key: str = 'SAMPLE FORMAT', data_format_key: str = 'DATA FORMAT',
                 sample_count_key: str = 'SAMPLES TOTAL'):
        """
        This function is the default constructor that loads the data from the filename that is provided in the argument
        list.  This will load the header and the acoustic data into a series of arrays and DataFrames.

        Parameters
        ----------
        filename : STRING, default = None
            This is the full path to the location of the file that we want to load into the class
        wfm: Waveform, default = None
            This is a Waveform object that we want to store in the standard binary file format
        header_element_count: int, default = None
            The number of elements within the header to read. This is typically the first line of the file,
            but it may or may not include the first line in the count. This provides a method to override the
            internal determination of the line count. The default is none, in which case the constructor assumes that
            information in the file is correct.
        sample_rate_key: str, default = "SAMPLE RATE (HZ)"
            This is the key to use in searching the header dictionary to define the number of samples per second the
            digital data was sampled during the measurement.
        start_time_key: str, default = "TIME (UTC ZULU)"
            This is the key to use in searching the header dictionary to define the start time of the
            measurement.
        sample_format_key: str, default = "SAMPLE FORMAT"
            This is the key to use in searching the header dictionary to define sample format for the binary data
        data_format_key: str, default = "DATA FORMAT"
            This is the key to use in searching the header dictionary to define the format of the sample stored in
            the binary data.
        sample_count_key: str, default = "SAMPLES TOTAL"
            The number of samples within the binary section of the file.

        Returns
        -------
        None.

        Remarks
        _______
        20230201 - FSM - Updated the methods to read the header information so that a variety of formatted
        information can be used in definition of the StandardBinaryFile object.
        """

        self.header = None

        if filename is not None:
            try:
                #   Open the file for reading in binary format
                f_in = open(filename, 'rb')

                #   Read the lines of header information
                name, value = StandardBinaryFile.convert_header(StandardBinaryFile.read_line(f_in))

                #   This is the header line, so now we can determine how many total lines of header information is
                #   present in the file
                if header_element_count is None:
                    self.header_line_count = int(value)
                elif header_element_count is not None:
                    self.header_line_count = header_element_count

                #   Read through the lines and extract the data as command and values that are inserted into a
                #   dictionary.
                self.header = dict()
                for i in range(self.header_line_count - 1):
                    #   Split the data in the header line
                    name, value = StandardBinaryFile.convert_header(StandardBinaryFile.read_line(f_in))

                    #   In effort to make the CSV representation of the data from the TimeHistory functions we need
                    #   to ensure that the commas and extra carriage return/line feeds are removed.
                    while ',' in name:
                        name = name.replace(',', ';')
                    while ',' in value:
                        value = value.replace(',', ';')

                    while '\r' in name:
                        name = name.replace('\r', '')

                    while '\n' in name:
                        name = name.replace('\n', '')

                    #   Assign the key and value within the dictionary
                    self.header[name] = value

                #   Now to effectively understand how to read the data from the binary portion, we must determine
                #   where specific data within the header exist. So look for the elements that were defined within
                #   the function prototype.
                #
                #   The sample rate
                if not (sample_rate_key in self.header.keys()):
                    raise ValueError("The name of the sample rate element of the waveform is not located within the "
                                     "header dictionary. Please provide the correct name of the sample rate property")
                else:
                    self.fs = float(self.header[sample_rate_key])

                #   The start time of the audio file
                if not (start_time_key in self.header.keys()):
                    raise ValueError(
                        "The name of the start time element of the waveform is not located within the "
                        "header dictionary. Please provide the correct name of the start time property")
                else:
                    self.time0 = parser.parse(self.header[start_time_key])

                #   The number of samples in the waveform
                if not (sample_count_key in self.header.keys()):
                    raise ValueError("The number of samples must be provided, and the expected header element is not "
                                     "found within the list of objects in the header.")
                else:
                    self.sample_count = int(self.header[sample_count_key])

                #   At this point there should be no reason for the data to be stored as anything other than REAL*4
                #   Little Endian, but we do not account for any other formats, so we must now examine what is in the
                #   header and exit if it is not what we expect.
                if not (sample_format_key in self.header.keys()):
                    raise ValueError(
                        "The name of the sample format element of the waveform is not located within the "
                        "header dictionary. Please provide the correct name of the sample format property")
                else:
                    if self.header[sample_format_key].upper() != "LITTLE ENDIAN":
                        raise ValueError("The expected format is not present in the header.")

                if not (data_format_key in self.header.keys()):
                    raise ValueError(
                        "The name of the data format element of the waveform is not located within the "
                        "header dictionary. Please provide the correct name of the data format property")
                else:
                    if self.header[data_format_key].upper() != "REAL*4":
                        raise ValueError("The required sample formate is not present in the header.")

                #   Read the data
                self._samples = np.asarray(struct.unpack('f' * self.sample_count, f_in.read(4 * self.sample_count)))

                #   close the file
                f_in.close()

            except IndexError:
                f_in.close()

                raise ValueError()
            except ValueError:
                f_in.close()

                raise ValueError()
        elif wfm is not None:
            super().__init__(wfm.samples, wfm.sample_rate, wfm.start_time)
            self.header = None
        else:
            self._samples = None
            self.fs = None
            self.time0 = None
            self.header = None

    @staticmethod
    def convert_header(header_line):
        """
        This function will take the information within the header line and remove
        the semicolon in the front and all elipsoid markers to determine the name
        of the property.  It also splits based on the colon to determine the value

        @author: Frank Mobley

        Parameters
        ----------
        header_line : STRING
            The line of text from the header of the file

        Returns
        -------
        name : STRING
            The name of the property or attribute

        value : STRING
            The value of the property

        """

        #   Split the string based on the colon

        elements = header_line.split(':')

        if len(elements) > 2:
            value = ':'.join(elements[1:])
        else:
            value = elements[1].strip()
        name = elements[0][1:].split('.')[0]

        return name, value

    @staticmethod
    def read_line(binary_file):
        """
        Python does not provide the ability to read a line of text from a
        binary file.  This function will read from the current position in the
        file to the new line character.  The set of bytes is then converted to
        a string and returned to the calling function.

        @author: frank Mobley

        Parameters
        ----------
        binary_file : FILE
            The file pointer that will be read from

        Returns
        -------
        The string representing a line of ASCII characters from the file.

        """

        #   Get the current position within the file so that we can return here
        #   after determining where the end of the file is.

        current_position = binary_file.tell()

        #   Find the end of the file

        binary_file.seek(-1, 2)

        eof = binary_file.tell()

        #   Return to the point we were within the file

        binary_file.seek(current_position, 0)

        #   Read until the last character is a new line or we have reached the
        #   end of the file.

        characters = ''
        char = ' '
        while ord(char) != 10 or binary_file.tell() == eof - 1:
            char = binary_file.read(1)
            if ord(char) != 10:
                characters += char.decode()

        return characters

    @staticmethod
    def write_std_binary_file(wfm, orig_header, output_filename):
        """
        Static method that given a generic_time_waveform object and a header dict writes a standard binary file
        according to ANSI S12.75
        Parameters
        ----------
        wfm : generic_time_waveform object
            Waveform of measured data.

        orig_header : dict
            Header info.

        output_filename : str
            Output file path with full name of file and extension.
        """
        # Check that correct data type were given as inputs
        if not (isinstance(wfm, Waveform)) or not (isinstance(orig_header, dict)) or \
                not (isinstance(output_filename, str)):

            raise ValueError("Incorrect inputs, check that wfm is a generic time waveform object, header is a dict and "
                             "output_filename is a str that's a full file path to the output directory")

        header = orig_header.copy()

        # Append sample format and encoding method if they dont exist to the header dict

        header['SAMPLE RATE (HZ)'] = int(np.floor((wfm.sample_rate)))
        header['SAMPLES TOTAL'] = len(wfm.samples)
        header['TIME (UTC ZULU)'] = wfm.start_time.strftime('%Y/%m/%d %H:%M:%S.%f')
        header['SAMPLE FORMAT'] = "LITTLE ENDIAN"
        header['DATA FORMAT'] = 'REAL*4'

        # Check to see if output path exist and open file if it doesn't exist

        if not os.path.exists(output_filename):
            f = open(output_filename, 'wb')

            # Write header info from dict

            header_line = ';{}'.format("HEADER SIZE").ljust(41, '.') + ': {}\n'.format(len(header.keys())+1)
            f.write(header_line.encode('utf-8'))

            for key in header.keys():
                header_line = ';{}'.format(key.upper()).ljust(41, '.') + ': {}\n'.format(header[key])
                f.write(header_line.encode('utf-8'))

            # Write pressure data to end of file

            for i in range(len(wfm.samples)):
                f.write(struct.pack('<f', wfm.samples[i]))
            f.close()

    def make_wav_file(self, wav_path, archival_location=None, artist=None, commissioning_organization=None,
                      copyright=None, cropping_information=None, originating_object_dimensions=None, engineer_name=None,
                      subject_genre=None, key_words=None, originating_object_medium=None, title=None, subject_name=None,
                      description=None, data_source=None, original_form=None, digitizing_engineer=None,
                      track_number=None):
        """
        The data within the header can be created with the LIST chunk that is common with audio files.  This requires
        the generation of a set of information within the meta_data element of the id3_wave_file that forms the
        information, which is typically at the end of the audio file.  This function converts the data within the
        StandardBinaryFile into a scaled version of the wave file with a LIST tag dataset.
        :param wav_path: str - the output of path for the file
        :param archival_location: str - the location of the original and all additional copies of the data
        :param artist: str - the person who is making the recording
        :param commissioning_organization: str - the organization that requested the measurement
        :param copyright: str - any description of the copyright and classification of the data within the file
        :param cropping_information: str - was the data cropped, is so what was the start and stop sample
        :param originating_object_dimensions: str - any information regarding the dimensions of the object that was
            recorded
        :param engineer_name : str - the name of the engineer that recorded the data
        :param subject_genre : str - the genre of the data
        :param key_words: str - a comma delimited collection of key words that describes this data
        :param originating_object_medium: str - what was the original medium that contained this data
        :param title: str - the title of the file
        :param subject_name: str - a name for the subject of the file
        :param description: str - further descriptions of the data within this file
        :param data_source: str - the name of the source of the data, which is different from the commissioned
            organization in that we want to know where the digital file came from
        :param original_form: str - a description of the original form that this data existed in before the wav file
            was created
        :param digitizing_engineer: str - the name/identifier of the individual who is creating this file - if not
            supplied the username of the individual logged into the computer running this code is used
        :param track_number: int - the track number for this wav file
        """

        #   get the username of the person running this code

        import getpass
        username = getpass.getuser()

        #   Create the wave_file

        wav = WaveFile()

        #   Add the header information to the class

        for key in self.header.keys():
            wav.header[key] = self.header[key]

        #   Now we need to put the generic_time_waveform information into the class object

        wav.samples = self.samples
        wav.start_time = self.time0
        wav.sample_rate = self.sample_rate

        #   Now we can create the specific tags within the common LIST chunk with information from the header

        if archival_location is not None:
            wav.archival_location = archival_location
        if artist is not None:
            wav.artist = artist
        if commissioning_organization is not None:
            wav.commissioning_organization = commissioning_organization
        if copyright is not None:
            wav.copyright = copyright

        wav.creation_date = datetime.now()

        if cropping_information is not None:
            wav.cropping_information = cropping_information
        if originating_object_dimensions is not None:
            wav.originating_object_dimensions = originating_object_dimensions
        if engineer_name is not None:
            wav.engineer_name = engineer_name
        else:
            wav.engineer_name = username
        if subject_genre is not None:
            wav.subject_genre = subject_genre
        if key_words is not None:
            wav.key_words = key_words
        if originating_object_medium is not None:
            wav.originating_object_medium = originating_object_medium
        if title is not None:
            wav.title = title
        else:
            wav.title = os.path.basename(wav_path)
        if subject_name is not None:
            wav.subject_name = subject_name
        if description is not None:
            wav.description = description

        wav.creation_software = "Python, AFRL_Physical_Acoustics "

        if data_source is not None:
            wav.data_source = data_source
        if original_form is not None:
            wav.original_form = original_form
        if digitizing_engineer is not None:
            wav.digitizing_engineer = digitizing_engineer
        else:
            wav.digitizing_engineer = username

        if track_number is not None:
            wav.track_number = track_number

        #   Save the data to the file

        if not(os.path.exists(os.path.dirname(wav_path))):
            os.makedirs(os.path.dirname(wav_path))

        wav.normalized = True
        wav.save(wav_path)





from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

    with open('HISTORY.md') as history_file:
        HISTORY = history_file.read()

        with open('LICENSE.txt') as license_file:
            LICENSE = license_file.read()

setup_args = dict(
    name='PyTimbre',
    version='0.5.16',
    description='Python conversion of Timbre Toolbox',
    long_description_content_type="text/markdown",
    long_description=README+'\n\n' + HISTORY+'\n\n' + LICENSE,
    license='MIT',
    packages=find_packages(),
    author='Dr. Frank Mobley',
    author_email='frank.mobley.1@afrl.af.mil',
    keywords=['machine learning', 'feature extraction', 'MATLAB', 'audio'],
    url='https://gitlab.com/python-audio-feature-extraction/pytimbre',
    download_url=''
)

install_requires = [
    'numpy',
    'scipy',
    'pyfftw',
    'statsmodels',
    'mosqito',
    'colorednoise'
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)


import warnings
import datetime
import numpy as np
import scipy.fft
import scipy.signal
from .fractional_octave_band import FractionalOctaveBandTools as fob_tools
from ..waveform import Waveform
from .acoustic_weights import AcousticWeights


class Spectrum:
    """
    This is the base class that defines the structure of the spectrum object. It does not calculate the frequency
    spectrum from a waveform, but can be used to represent a single spectrum that was previously created.

    Remarks
    2022-12-13 - FSM - Added a function to calculate the sound quality metrics based on the frequency spectrum

    """
    def __init__(self, a: Waveform = None):
        """
        The default constructor that builds the information within the Spectrum class based on the contents of the
        object "a".

        Parameters
        ----------
        a: Waveform - the acoustic samples that define the source of the time-domain information that we are interested
        in processing
        """

        self._waveform = a
        self._frequencies = None
        self._acoustic_pressures_pascals = None
        self._time0 = None

        self._bandwidth = None
        self._f0 = None
        self._f1 = None

        if a is not None:
            self._time0 = self.waveform.start_time

            if len(a.samples.shape) > 1:
                self._waveform.samples = self._waveform.samples.reshape((-1,))

        self._fft_size = 4096
        self._window_size_seconds = 0.0232
        self._hop_size_seconds = 0.0058
        self._window_size = self._window_size_seconds * self.sample_rate
        self._hop_size = self.hop_size_seconds * self.sample_rate

        self.bin_size = self.sample_rate / self._fft_size
        self.sample_rate_x = self.sample_rate / self.hop_size
        self.sample_rate_y = self._fft_size / self.sample_rate_x
        self.window = np.hamming(self.window_size)
        self.window_overlap = self.window_size - self.hop_size

        #   Define the features

        self.centroid = None
        self._mean_center = None
        self.spread = None
        self.skewness = None
        self.kurtosis = None
        self.slope = None
        self.decrease = None
        self.roll_off = None
        self.energy = None
        self.flatness = None
        self.crest = None
        self.probability_distribution = None
        self.integration_variable = None
        self.geometric_mean = None
        self.arithmetic_mean = None

    #   ------------------------------------------ Protected functions -------------------------------------------------

    def _calculate_spectrum(self):
        import warnings

        warnings.warn("This function does nothing. You will need to implement this function in the child class to build"
                      "the spectrum with the correct units and format.")

        raise ValueError('This method must be implemented and assigned frequencies within the calculation '
                         'of each type of spectrum class.')

    #   ------------------------------------- Class properties ---------------------------------------------------------

    @property
    def waveform(self):
        return self._waveform

    @property
    def signal(self):
        return self.waveform.samples

    @property
    def sample_rate(self):
        if self._waveform is None:
            return 48000
        else:
            return self.waveform.sample_rate

    @property
    def duration(self):
        if self._waveform is None:
            raise AttributeError("No waveform has been provided to the Spectrum object.")
        return self.waveform.duration

    @property
    def time(self):
        if self._time0 is None:
            raise AttributeError("No time object has been provided to the Spectrum object.")
        return self._time0

    @property
    def time_past_midnight(self):
        if self._time0 is None:
            raise AttributeError("No time has been provided to the Spectrum object.")

        if isinstance(self._time0, datetime.datetime):
            return 60 * (60 * self._time0.hour + self._time0.minute) + self._time0.second + \
                   float(self._time0.microsecond / 1e6)
        else:
            return self._time0

    @property
    def frequencies(self):
        if self._frequencies is None:
            self._calculate_spectrum()

        return self._frequencies

    @frequencies.setter
    def frequencies(self, values):
        self._frequencies = values

    @property
    def pressures_pascals(self):
        if self._frequencies is None or self._acoustic_pressures_pascals is None:
            self._calculate_spectrum()

        return self._acoustic_pressures_pascals

    @pressures_pascals.setter
    def pressures_pascals(self, values):
        self._acoustic_pressures_pascals = values

    @property
    def pressures_decibels(self):
        return 20 * np.log10(self.pressures_pascals / 20e-6)

    @property
    def overall_level(self):
        """
        Overall sound pressure level, unweighted (i.e. flat weighted, Z-weighted).  Calculated as the energetic sum
        of the power spectrum.
        """
        return AcousticWeights.lf(self.pressures_decibels)

    @property
    def overall_a_weighted_level(self):
        """
        A-weighted overall sound pressure level.  Calculated as the energetic sum
        of the A-weighted power spectrum.
        """

        return AcousticWeights.la(self.pressures_decibels, self.frequencies)

    @property
    def fractional_octave_bandwidth(self):
        return self._bandwidth

    @fractional_octave_bandwidth.setter
    def fractional_octave_bandwidth(self, value):
        self._bandwidth = value

    @property
    def start_fractional_octave_frequency(self):
        return self._f0

    @start_fractional_octave_frequency.setter
    def start_fractional_octave_frequency(self, value):
        self._f0 = value

    @property
    def stop_fractional_octave_frequency(self):
        return self._f1

    @stop_fractional_octave_frequency.setter
    def stop_fractional_octave_frequency(self, value):
        self._f1 = value

    @property
    def narrowband_frequency_count(self):
        return self._fft_size

    @narrowband_frequency_count.setter
    def narrowband_frequency_count(self, value):
        self._fft_size = value

        self.bin_size = self.sample_rate / self._fft_size
        self.sample_rate_x = self.sample_rate / self.hop_size
        self.sample_rate_y = self._fft_size / self.sample_rate_x
        self.window = np.hamming(self.window_size)
        self.window_overlap = self.window_size - self.hop_size

    @property
    def roughness(self):
        from mosqito.sq_metrics import roughness_dw_freq

        return roughness_dw_freq(spectrum=self.pressures_decibels, freqs=self.frequencies)[0]

    @property
    def loudness(self):
        from mosqito.sq_metrics import loudness_zwst_freq

        return loudness_zwst_freq(self.pressures_decibels, self.frequencies)

    @property
    def sharpness(self):
        from mosqito.sq_metrics import sharpness_din_freq

        return sharpness_din_freq(self.pressures_decibels, self.frequencies)

    @property
    def hop_size_seconds(self):
        return self._hop_size_seconds

    @hop_size_seconds.setter
    def hop_size_seconds(self, value: float):
        self._hop_size_seconds = value
        self.hop_size = int(np.floor(self.hop_size_seconds * self.sample_rate))
        self.sample_rate_x = self.sample_rate / self.hop_size
        self.window_overlap = self.window_size - self.hop_size

    @property
    def window_size(self):
        return int(np.floor(self._window_size))

    @window_size.setter
    def window_size(self, value: int):
        self._window_size = value
        self._window_size_seconds = self.window_size / self.sample_rate
        self.window = np.hamming(self.window_size)
        self.window_overlap = self.window_size - self.hop_size

    @property
    def hop_size(self):
        return int(np.floor(self._hop_size))

    @hop_size.setter
    def hop_size(self, value: int):
        self._hop_size = value
        self.hop_size_seconds = self._hop_size / self.sample_rate
        self.sample_rate_x = self.sample_rate / self._hop_size
        self.window_overlap = self.window_size - self._hop_size

    @property
    def spectral_centroid(self):
        """
        Spectral centroid represents the spectral center of gravity.
        """

        if self.centroid is None:
            if self._acoustic_pressures_pascals is None and self._frequencies is None:
                self._calculate_spectrum()

            if self.probability_distribution is None or self.integration_variable is None:
                self.calculate_normalized_distribution()

            self.centroid = np.sum(self.integration_variable * self.probability_distribution, axis=0)

        return self.centroid

    @property
    def spectral_spread(self):
        """
        Spectral spread or spectral standard-deviation represents the spread of the spectrum around its mean value.
        """
        if self.mean_center is None:
            self._calculate_mean_center()

        if self.probability_distribution is None or self.integration_variable is None:
            self.calculate_normalized_distribution()

        if self.spread is None:
            self.spread = np.sqrt(np.sum(self.mean_center ** 2 * self.probability_distribution, axis=0))

        return self.spread

    @property
    def spectral_skewness(self):
        """
        Spectral skewness gives a measure of the asymmetry of the spectrum around its mean value. A value of 0 indicates
        a symmetric distribution, a value < 0 more energy at frequencies lower than the mean value, and values > 0 more
        energy at higher frequencies.
        """
        if self.mean_center is None:
            self._calculate_mean_center()

        if self.probability_distribution is None or self.integration_variable is None:
            self.calculate_normalized_distribution()

        if self.skewness is None:
            self.skewness = np.sum(self.mean_center ** 3 * self.probability_distribution, axis=0) / \
                            self.spectral_spread ** 3

        return self.skewness

    @property
    def spectral_kurtosis(self):
        """
        Spectral kurtosis gives a measure of the flatness of the spectrum around its mean value. Values approximately 3
        indicate a normal (Gaussian) distribution, values less than 3 indicate a flatter distributions, and values
        greater than 3 indicate a peakier distribution.
        """
        if self.mean_center is None:
            self._calculate_mean_center()

        if self.probability_distribution is None or self.integration_variable is None:
            self.calculate_normalized_distribution()

        if self.kurtosis is None:
            self.kurtosis = np.sum(self.mean_center ** 4 * self.probability_distribution, axis=0) / \
                            self.spectral_spread ** 4

        return self.kurtosis

    @property
    def spectral_slope(self):
        """
        Spectral slope is computed using a linear regression over the spectral amplitude values. It should be noted that
        the spectral slope is linearly dependent on the spectral centroid.
        """
        if self._acoustic_pressures_pascals is None and self._frequencies is None:
            self._calculate_spectrum()

        if self.probability_distribution is None or self.integration_variable is None:
            self.calculate_normalized_distribution()

        if self.slope is None:
            numerator = len(self._frequencies) * (self._frequencies.transpose().dot(self.probability_distribution))
            numerator -= np.sum(self._frequencies) * np.sum(self.probability_distribution, axis=0)
            denominator = len(self._frequencies) * sum(self._frequencies ** 2) - np.sum(self._frequencies) ** 2
            self.slope = numerator / denominator

        return self.slope

    @property
    def spectral_decrease(self):
        """
        Spectral decrease was proposed by Krimphoff (1993) in relation to perceptual studies. It averages the set of
        slopes between frequency f[k] and f[1]. It therefore emphasizes the slopes of the lowest frequencies.
        """
        if self._acoustic_pressures_pascals is None and self._frequencies is None:
            self._calculate_spectrum()

        if self.probability_distribution is None or self.integration_variable is None:
            self.calculate_normalized_distribution()

        if self.decrease is None:
            numerator = self._acoustic_pressures_pascals[1:] - self._acoustic_pressures_pascals[0]
            denominator = (1 / np.arange(1, len(self._frequencies)))
            self.decrease = (denominator.dot(numerator)).transpose().reshape((-1,))
            self.decrease /= np.sum(self.probability_distribution[1:], axis=0)

        return self.decrease

    @property
    def spectral_roll_off(self):
        """
        Spectral roll-off was proposed by Scheirer and Slaney (1997). It is defined as the frequency below which 95%
        of the signal energy is contained. The value is returned as the normalized frequency (i.e. you must multiply
        by the sample rate to determine the actual frequency of the roll-off.
        """
        if self._acoustic_pressures_pascals is None and self._frequencies is None:
            self._calculate_spectrum()

        if self.roll_off is None:
            threshold = 0.95
            cum_sum = np.cumsum(self._acoustic_pressures_pascals, axis=0)
            sum = np.ones((len(self.frequencies), )) * (threshold * np.sum(self._acoustic_pressures_pascals))

            bin = np.cumsum(1 * (cum_sum > sum), axis=0)
            idx = np.where(bin == 1)[0]

            self.roll_off = self.frequencies[idx][0]

        return self.roll_off

    @property
    def spectral_energy(self):
        """
        A summation of the energy within the spectrum
        """
        if self._acoustic_pressures_pascals is None and self._frequencies is None:
            self._calculate_spectrum()

        if self.energy is None:
            self.energy = np.sum(self._acoustic_pressures_pascals, axis=0)

        return self.energy

    @property
    def spectral_flatness(self):
        """
        Spectral flatness is obtained by comparing the geometrical mean and the arithmetical mean of the spectrum. The
        original formulation first splot the spectrum into various frequency bands (Johnston, 1988). However, in the
        context of timbre characterization, we use a single frequency band covering the whole frequency range. For
        tonal signals, the spectral flatness is close to 0( a peaky spectrum), whereas for noisy signals it is close to
        1 (flat spectrum).
        """
        if self._acoustic_pressures_pascals is None and self._frequencies is None:
            self._calculate_spectrum()

        if self.flatness is None:
            self.geometric_mean = np.exp((1 / len(self._frequencies)) * np.sum(np.log(self._acoustic_pressures_pascals), axis=0))
            self.arithmetic_mean = np.mean(self._acoustic_pressures_pascals, axis=0)
            self.flatness = self.geometric_mean / self.arithmetic_mean

        return self.flatness

    @property
    def spectral_crest(self):
        """
        The spectral crest measure is obtained by comparing the maximum value and arithmetical mean of the spectrum.
        """
        if self._acoustic_pressures_pascals is None and self._frequencies is None:
            self._calculate_spectrum()

        if self.arithmetic_mean is None:
            self.arithmetic_mean = np.mean(self._acoustic_pressures_pascals, axis=0)

        if self.crest is None:
            self.crest = np.max(self._acoustic_pressures_pascals, axis=0) / self.arithmetic_mean

        return self.crest

    @property
    def mean_center(self):
        if self._mean_center is None:
            self._calculate_mean_center()

        return self._mean_center

    def calculate_engineering_unit_scale_factor(self, calibration_level: float = 94, calibration_frequency=1000):
        """
        This will take the data within the class and build the spectral time history and then determine the value of the
        scale factor to get a specific sound pressure level at a certain frequency.

        Parameters
        ----------
        calibration_level: float - the value of the acoustic level for the calibration
        calibration_frequency: float - the value of the frequency that is supposed to be used for the calibration

        Returns
        -------
        float - the engineering scale factor that can be directly applied to the acoustic data
        """

        if self.fractional_octave_bandwidth is None:
            raise ValueError("This analysis cannot be accomplished on a narrowband spectrum")
        elif self.fractional_octave_bandwidth != 3:
            raise ValueError("This analysis must be accomplished on a one-third-octave spectrum")
        elif self.start_fractional_octave_frequency != 10:
            raise ValueError("Until the algorithm is adjusted, this analysis requires the spectrum to be defined "
                             "starting at 10 Hz")

        #   Now determine the index of the band within the spectral time history that should be examined for the
        #   calculation of the engineering scaling units.

        idx = int(np.floor(fob_tools.nearest_band(3, calibration_frequency))) - 10

        #   Now this may select the nearest band as the band just above the actual band.  So ensure that the lower band
        #   is below the desired center frequency

        if fob_tools.lower_frequency(3, idx + 10) > calibration_frequency:
            idx -= 1

        #   From the spectrum obtain the values of the frequency

        calibration_values = self.pressures_decibels[idx]

        #   Calculate the sensitivity of each time history spectra

        sens = calibration_level - calibration_values
        sens /= 20
        sens *= -1
        sens = 10.0 ** sens

        return sens

    def get_average_features(self, include_sq_metrics: bool = True):
        import numpy as np
        """
        This will return a dict of the various elements within the spectrum and waveform (if it was used to create the
        spectrogram object) with any time variant elements averaged.
        """

        features = dict()

        if self.waveform is not None:
            features = self.waveform.get_features(include_sq_metrics)
            features['zero_crossing_rate'] = np.mean(features['zero crossing rate'])
            auto_correlation_coefficients = np.mean(features['auto-correlation'], axis=0)
            if include_sq_metrics:
                features['loudness'] = np.mean(features['loudness'])
                features['sharpness'] = np.mean(features['sharpness'])
                features['roughness'] = np.mean(features['roughness'])

            for i in range(12):
                features['auto_correlation_{:02.0f}'.format(i)] = auto_correlation_coefficients[i]

            del features['zero crossing rate']
            del features['auto-correlation']

        features['spectral_centroid'] = np.mean(self.spectral_centroid)
        features['spectral_spread'] = np.mean(self.spectral_spread)
        features['spectral_skewness'] = np.mean(self.spectral_skewness)
        features['spectral_kurtosis'] = np.mean(self.spectral_kurtosis)
        features['spectral_slope'] = np.mean(self.spectral_slope)
        features['spectral_decrease'] = np.mean(self.spectral_decrease)
        features['spectral_roll_off'] = np.mean(self.spectral_roll_off)
        features['spectral_energy'] = np.mean(self.spectral_energy)
        features['spectral_flatness'] = np.mean(self.spectral_flatness)
        features['spectral_crest'] = np.mean(self.spectral_crest)

        return features

    def calculate_normalized_distribution(self):
        if self._acoustic_pressures_pascals is None:
            self._calculate_spectrum()

        self.probability_distribution = self._acoustic_pressures_pascals
        self.probability_distribution /= np.sum(self._acoustic_pressures_pascals, axis=0)

        self.integration_variable = self.frequencies

    def _calculate_mean_center(self):
        if self._acoustic_pressures_pascals is None:
            self._calculate_spectrum()

        if self.probability_distribution is None or self.integration_variable is None:
            self.calculate_normalized_distribution()

        if self._mean_center is None:
            self._mean_center = self.integration_variable - self.spectral_centroid

        return self._mean_center


class SpectrumByFFT(Spectrum):
    """
    This class is a specialization of the spectrum class that implements the way to define the spectrum. This is
    accomplished with the Fourier Transform, using a chunked overlapping representation of the data within the
    calculation.

    In this class the frequencies property represents the single-sided frequencies of the Fourier transform
    """

    def __init__(self, a: Waveform = None, fft_size: int = None):
        """
        Constructor - This constructor adds the ability of the class to define the specific size of the Fourier
        frequency bins. This also adds a number of elements that represent the double and single sided frequencies
        that originate with the Fourier transform methodologies.
        """
        #   Call the base constructor to define the waveform within the class
        super().__init__(a)

        #   Now define some information within the class related to the frequencies and the information generated
        #   specifically for the FFT methods
        self._fft_size = fft_size
        self._frequencies_double_sided = None
        self._frequencies_nb = None

        #   Define some other representations of the acoustic information that might be needed in various
        #   analyses.
        self._pressures_double_sided_complex = None

        if a is not None:
            #   There is no default value for the FFT size, so let's do some analysis to determine what the most optimal
            #   value of this should be if it is not provided.
            if self._fft_size is None:

                #   Set the default block size
                self._fft_size = int(2 ** np.floor(np.log2(len(self.waveform.samples))))

            elif self._fft_size > len(self.waveform.samples):
                raise ValueError('FFT block size cannot be greater than the total length of the signal.')

    def _calculate_spectrum(self):
        """
        This function generates the complex spectra using an FFT of multiple blocks of an input waveform,
        where the blocks have 50% overlap and are each weighted by a Hanning window.

        The FFT blocks are then scaled to contain the appropriate amount of energy for input into a power
        spectrum calculation.
        """

        #   Create the frequency arrays
        self._frequencies_double_sided = self.sample_rate * np.arange(0, self._fft_size) / self._fft_size
        self._frequencies_nb = self._frequencies_double_sided[:int(self._fft_size / 2)]
        df = self._frequencies_nb[1] - self._frequencies_nb[0]

        #   enforce a zero mean value
        x = self.waveform.samples - np.mean(self.waveform.samples)

        #   Generate a Hanning window
        ww = np.hanning(self._fft_size)
        W = np.mean(ww ** 2)

        #   Divide the total data into blocks with 50% overlap and Hanning window
        blocks = np.zeros(shape=(int(np.floor(2 * len(x) / self._fft_size - 1)), self._fft_size))

        for k in range(blocks.shape[0]):
            i = int(k * self._fft_size / 2)
            j = i + blocks.shape[1]
            blocks[k, :] = ww * x[i:j]

        #   Determine complex pressure amplitude
        self._pressures_double_sided_complex = np.sqrt(2 * df / self._fft_size /
                                                       self.sample_rate / W) * scipy.fft.fft(blocks, n=self._fft_size)

        #   Now assign the values for the acoustic pressures using this information, but only using the first hold of
        #   the frequency data.
        self._frequencies = self._frequencies_nb
        self._acoustic_pressures_pascals = self.auto_correlation_spectrum

    @property
    def frequency_increment(self):
        return self.frequencies[1] - self.frequencies[0]

    @property
    def frequencies_double_sided(self):
        """
        2-D Numpy array of double-sided frequencies from a Fourier transform of each block of a waveform
        with 50% overlap and a Hanning window in units of Pascals.
        """

        if self._frequencies_double_sided is None:
            self._calculate_spectrum()

        return self._frequencies_double_sided

    @property
    def pressures_complex_double_sided(self):
        """
        2-D Numpy array of double-sided complex pressures from a Fourier transform of each block of a waveform
        with 50% overlap and a Hanning window in units of Pascals.
        """
        if self._pressures_double_sided_complex is None:
            self._calculate_spectrum()

        return self._pressures_double_sided_complex

    @property
    def auto_correlation_spectrum(self):
        """
        Numpy array of single-sided real-valued pressures averaged over FFT of all blocks of a waveform with 50%
        overlap and a Hanning window.  Units of Pascals.
        """
        if self._pressures_double_sided_complex is None:
            self._calculate_spectrum()

        pressures_single_sided = self._pressures_double_sided_complex[:, :len(self._frequencies_nb)]
        return np.sqrt(np.mean((pressures_single_sided * np.conj(pressures_single_sided)).real, axis=0))

    @property
    def fft_size(self):
        return self._fft_size

    @property
    def power_spectral_density(self):
        """
        Numpy array of single-sided real-values. Pressures scaled by frequency, in units of Pascals / sqrt(Hz).
        """
        return self.pressures_pascals / np.sqrt(self.frequency_increment)

    @staticmethod
    def convert_nb_to_fob(frequencies_nb, pressures_pascals_nb, fob_band_width: int = 3, f0: float = 10,
                          f1: float = 10000):
        """
        This function converts the frequency and pressure arrays sampled in narrowband resolution to the fractional
        octave band resolution.

        Parameters
        ----------
        frequencies_nb: nd_array - the collection of narrowband frequencies that we want to convert
        pressures_pascals_nb: nd_array - the collection of pressures in units of pascals for the frequencies
        fob_band_width: int, default = 3 - the fractional octave band resolution that we desire
        f0: float, default = 10 - The start frequency of the output fractional octave frequencies
        f1: float, default = 10000 - The end frequency of the output fractional octave frequencies

        Returns
        -------
        frequencies_fob: nd_array - the collection of frequencies from f0 to f1 with the resolution of fob_band_width
        pressures_pascals_fob: nd_array - the fractional octave pressure in pascals at the associated frequencies
        """

        frequencies_fob = fob_tools.get_frequency_array(fob_band_width, f0, f1)

        #   Build the array of pressures that are the same size as the list of frequencies
        pressures_pascals_fob = np.zeros((len(frequencies_fob),))

        for i in range(len(frequencies_fob)):
            pressures_pascals_fob[i] = np.sqrt(sum(pressures_pascals_nb ** 2 *
                                                   fob_tools.filter_shape(fob_band_width, frequencies_fob[i],
                                                                          frequencies_nb)))

        return frequencies_fob, pressures_pascals_fob

    def to_fractional_octave_band(self, bandwidth: int = 3, f0: float = 10, f1: float = 10000):
        """
        This function will convert the spectrum from a narrowband resolution to a factional octave band resolution by
        applying the shape functions to the narrowband spectral values and determining the weighted value within the
        fractional octave band.

        Parameters
        ----------
        bandwidth: float, default = 3 - the fractional octave resolution that we will sample the frequency spectrum
        f0: float, default = 10 - the lowest frequency within the spectrum
        f1: float, default = 10000 - the heighest frequency within the spectrum

        Returns
        -------
        Spectrum - a spectrum object with the frequencies at the specified resolution and between the specified
        frequency values.
        """

        f_fob, p_fob = SpectrumByFFT.convert_nb_to_fob(self.frequencies, self.pressures_pascals, bandwidth, f0, f1)
        s = Spectrum()
        s.frequencies = f_fob
        s.pressures_pascals = p_fob
        s.start_fractional_octave_frequency = f0
        s.stop_fractional_octave_frequency = f1
        s.fractional_octave_bandwidth = bandwidth

        return s


class SpectrumByDigitalFilters(Spectrum):

    """
    This class differs from the previous classes in that the determination of the spectrum is accomplished through an
    application of digital filters. A set of filters describing the collection of filters at the highest desired full
    octave band will be constructed. These coefficients will be applied as the signal is recursively down sampled at
    half the sample rate. The minimum number of elements required to generate the filtered data require three times
    the number of coefficients, doubled until we reach the maximum sample rate of the system.

    Development
    20221008 - FSM - Since we added the bandwidth, start and stop frequencies to the base class, this class no longer
        required a separate definition. All references were adjusted to use the base class values for these parameters.
    """

    def __init__(self, a: Waveform, fob_band_width: int = 3, f0: float = 10, f1: float = 10000):
        """
        The constructor - this will determine what the minimum number of samples required is to generate the lowest
        desired frequency.

        Parameters
        ----------
        a: Waveform - the audio signal that we want to process
        fob_band_width: int, default: 3 - the bandwidth of the fractional octave
        f0: float, default: 10 - the lowest frequency in Hz
        f1: float, default: 10000 - the highest frequency in the output
        """
        super().__init__(a)

        self._bandwidth = fob_band_width
        self._f0 = f0
        self._f1 = f1
        self._b_coefficients = list()
        self._a_coefficients = list()

        #   Build the filters for the highest desired full octave
        #
        #   Determine the center frequency of the highest octave
        full_band = int(np.floor(fob_tools.nearest_band(1, self.stop_fractional_octave_frequency)))
        f_full = fob_tools.center_frequency(1, full_band)
        f_lo = fob_tools.lower_frequency(1, full_band)
        f_hi = fob_tools.upper_frequency(1, full_band)

        #   Now that the know the center frequency of the highest band, determine the upper limit, and then the closest
        #   center frequency in the desired bandwidth
        f_band = int(np.floor(fob_tools.nearest_band(self.fractional_octave_bandwidth, f_hi)))
        nyquist = self.sample_rate / 2.0

        #   Loop through the frequencies within the highest octave band and create the associated digital filters for
        #   the element based on the calculated high and low frequencies.
        while fob_tools.lower_frequency(self.fractional_octave_bandwidth, f_band) >= f_lo * 0.90:
            #   Define the window for the bandpass filter
            upper = fob_tools.upper_frequency(self.fractional_octave_bandwidth, f_band)
            lower = fob_tools.lower_frequency(self.fractional_octave_bandwidth, f_band)
            window = np.array([lower, upper]) / nyquist

            #   Create the filter coefficients for this frequency band and add it to the list for each coefficient set
            b, a = scipy.signal.butter(
                3,
                window,
                btype='bandpass',
                analog=False,
                output='ba'
            )

            self._b_coefficients.append(b)
            self._a_coefficients.append(a)

            #   Decrement the band number to move to the next band down.
            f_band -= 1

        #   Convert the lists to arrays
        self._b_coefficients = np.asarray(self._b_coefficients)
        self._a_coefficients = np.asarray(self._a_coefficients)

    @property
    def settle_time(self):
        return self.settle_samples / self.sample_rate

    @property
    def settle_samples(self):
        """
        Based on requirements of Matlab filtering, you must have at least 3 times the number of coefficients to
        accurately filter data. So this will start with that minimum, and then move through the full octave frequency
        band numbers to determine the minimum number of samples that are required for the filter to adequately settle.
        """
        #   Determine the band number for the lowest band
        low_band = int(np.floor(fob_tools.nearest_band(1, self.start_fractional_octave_frequency)))
        hi_band = int(np.floor(fob_tools.nearest_band(1, self.stop_fractional_octave_frequency)))

        minimum_required_points = 3 * self._b_coefficients.shape[1]

        for band_index in range(low_band + 1, hi_band + 1):
            minimum_required_points *= 2

        return minimum_required_points

    def _calculate_spectrum(self):
        """
        This will take the waveform that exist within the class and calculate the fractional octave pressures within
        each band that is adequately covered by the length of the waveform.
        """

        #   Create the list that will hold the frequencies and band pressures
        pressures = list()
        frequency = list()

        #   Determine the octave bands that will need to be calculated to cover the desired frequency range.
        low_band = int(np.floor(fob_tools.nearest_band(1, self.start_fractional_octave_frequency)))
        hi_band = int(np.floor(fob_tools.nearest_band(1, self.stop_fractional_octave_frequency)))

        #   Get the index of the band at the top of the full octave filter
        fob_band_index = int(np.floor(fob_tools.nearest_band(self.fractional_octave_bandwidth,
                                                             fob_tools.upper_frequency(1, hi_band))))

        #   Make a copy of the waveform that can be decimated
        wfm = Waveform(pressures=self.waveform.samples.copy(),
                       sample_rate=self.sample_rate,
                       start_time=self.waveform.start_time)

        #   Loop through the frequencies in reverse order
        for band_index in range(hi_band, low_band - 1, -1):
            #   if there are insufficient number of points in the waveform, terminate the process now
            if len(wfm.samples) < 3 * self._b_coefficients.shape[1]:
                warnings.warn("The number of points within the Waveform are insufficient to calculate digital filters "
                              "lower than these frequencies")
                break

            #   Now loop through the filter definitions that are presented in decreasing frequency magnitude
            for filter_index in range(self._b_coefficients.shape[0]):
                filtered_waveform = wfm.apply_iir_filter(self._b_coefficients[filter_index, :],
                                                         self._a_coefficients[filter_index, :])

                frequency.append(fob_tools.center_frequency(self.fractional_octave_bandwidth, fob_band_index))
                pressures.append(np.std(filtered_waveform.samples))

                fob_band_index -= 1

            #   Decimate the waveform, halving the sample rate and making the filter definitions move down a full octave
            if len(wfm.samples) / 2 < 3 * self._b_coefficients.shape[1]:
                warnings.warn("The number of points within the Waveform are insufficient to calculate digital filters "
                              "lower than these frequencies")

                break

            wfm = Waveform(pressures=scipy.signal.decimate(wfm.samples, 2),
                           sample_rate=wfm.sample_rate,
                           start_time=wfm.start_time)

        #   Convert the information within the pressures and frequency arrays into the correct elements for the class
        frequency = np.asarray(frequency)[::-1]
        pressures = np.asarray(pressures)[::-1]

        idx0 = np.where(frequency > fob_tools.lower_frequency(self.fractional_octave_bandwidth,
                                                              fob_tools.nearest_band(
                                                                  self.fractional_octave_bandwidth,
                                                                  self.start_fractional_octave_frequency)))[0][0]
        idx1 = np.where(frequency < fob_tools.upper_frequency(self.fractional_octave_bandwidth,
                                                              fob_tools.nearest_band(
                                                                  self.fractional_octave_bandwidth,
                                                                  self.stop_fractional_octave_frequency)))[0][-1]
        self._frequencies = frequency[np.arange(idx0, idx1 + 1)]
        self._acoustic_pressures_pascals = pressures[np.arange(idx0, idx1 + 1)]





import operator
import pandas
from .time_histories import TimeHistory, Spectrum, Waveform, WaveFile
import numpy as np


class Spectrogram (TimeHistory):
    """
    This class represents the Short Time Fourier Transform method for calculating the spectral variations with time. The
    code calculates the spectrogram using the same methods present in the STFT representation of the original Matlab 
    code.
    """

    def __init__(self, a: Waveform = None, integration_time: float = 0.25, nfft=4096):
        """
        Construct the information within the class from the waveform that is passed into the constructor
        
        :param a: Waveform - the acoustic information that we are interested in using for the timbre analysis.
        :param nfft: int - the resolution of the frequency analysis
        """
        super().__init__(a, integration_time)

        self._waveform = a

        self._levels = None
        self._frequencies = None
        self._times = None
        if a is not None:
            self._sample_rate = self._waveform.sample_rate
        else:
            self._sample_rate = 48000

        self._fft_size = nfft

        self._window_size_seconds = 0.0232
        self._hop_size_seconds = 0.0058
        self._window_size = self._window_size_seconds * self.sample_rate
        self._hop_size = self._hop_size_seconds * self.sample_rate

    # -------------------------------------------- Properties ----------------------------------------------------------

    @property
    def fft_size(self):
        return self._fft_size

    # -------------------------------------------- Protected Methods----------------------------------------------------

    def _calculate_spectrogram(self):
        import scipy.signal
        import scipy.fft

        #   If the window is centered at t, this is the starting index at which to loop up the signal which you want
        #   to multiply by the window.  It is a negative number because (almost) half of the window will be before
        #   time t and half after.  In fact, if the length of the window N is an even number, it is set up so this
        #   number equals -1 * (N / 2 -1).  If the length of the window is od, this number equals -1 * (N - 1) / 2
        left_hand_window_size = int(np.ceil(-(self._window_size - 1) / 2))

        #   This is the last index at which to look up signal values and is equal to (N - 1) / 2 if the length N of
        #   the window is odd and N / 2 if the length of the window is even.  This means that in the even case, the
        #   window has an unequal number of past and future values, i.e., time t is not the center of the window,
        #   but slightly to the left of the center of the window (before it).
        right_hand_window_size = int(np.ceil((self._window_size - 1) / 2))

        #   pre-pad the signal and calculate the hilbert transform
        signal = np.concatenate((np.zeros((-left_hand_window_size,)), self.signal))
        signal = scipy.signal.hilbert(signal)
        last_index = np.floor((len(self.signal) -
                               (right_hand_window_size + 1)) / self._hop_size) * self._hop_size + 1

        #   Define some support vectors
        index = np.arange(0, last_index, self._hop_size, dtype=int) - left_hand_window_size
        size_x = len(index)
        size_y = self.fft_size / 2
        self._times = np.arange(0, size_x) / (self.sample_rate / self._hop_size)
        normalized_frequency = np.arange(0, size_y) / size_y / 2

        #   Create the windowed signal
        window = np.hamming(int(self._window_size + 1))
        distribution_pts = np.zeros((self.fft_size, size_x), dtype='complex')
        for i in range(size_x):
            rng = np.arange(0, int(self._window_size + 1), dtype=int) + (index[i] + left_hand_window_size)

            distribution_pts[:int(self._window_size + 1), i] = signal[rng] * window

        #   Calculate the FFT
        distribution_pts = scipy.fft.fft(distribution_pts, n=self.fft_size, axis=0)

        #   Apply the specific scaling for the analysis
        distribution_pts = abs(distribution_pts)
        distribution_pts /= np.sum(abs(window))

        #   Only keep the first half of the spectrum
        levels = distribution_pts[:int(round(self.fft_size / 2)), :]

        #   Now build the collection of Spectrum objects.
        self._spectra = np.empty((len(self._times),), dtype=Spectrum)

        for i in range(len(self._times)):
            s = Spectrum()
            #   TODO: Insert the distribution_pts into the Spectrum class?
            s.frequencies = normalized_frequency
            s.pressures_pascals = levels[:, i]
            s._time0 = self._times[i]
            #s._waveform = distribution_pts[:, i]

            self._spectra[i] = s

    @staticmethod
    def from_data(levels, frequencies, times):
        """
        This function constructs the Spectrogram object from information obtained from the users and sets up an object
        that can be compared with external data without concern for differences in the methods to calculate the
        spectrogram data.

        :param levels: array-like - the 2-D levels with shape = [len(times), len(frequencies)]
        :param frequencies: array-like - the collection of frequencies that define one dimension of the levels matrix
        :param times: array-like - the collection of times within the spectrogram that define the second dimension
        :returns: Spectrogram object
        """

        s = Spectrogram()

        s._spectra = np.empty((len(times),), dtype=Spectrum)

        for i in range(len(times)):
            spec = Spectrum()
            spec._time_past_midnight = times[i]
            spec.frequencies = frequencies
            spec.pressures_pascals = levels[:, i]

            s._spectra[i] = spec

        return s
